#!/usr/bin/env python3
# coding: utf-8

"""
Functional tests for 'annotate' subcommand
"""

from PanACoTA.subcommands import annotate as annot
import test.test_unit.utilities_for_tests as tutil

import pytest
import os
import subprocess
import shutil
import time
import argparse
import matplotlib
import logging
import glob
matplotlib.use('AGG')


# LOGFILE_BASE = "test_main_from_parse"
# Define variables used by several tests
DBDIR = os.path.join("test", "data", "annotate")
GEN_PATH = os.path.join(DBDIR, "genomes")
EXP_DIR = os.path.join(DBDIR, 'exp_files')
TEST_DIR = os.path.join(DBDIR, 'test_files')
GENEPATH = os.path.join(DBDIR, "generated_by_func-tests")


@pytest.fixture(autouse=True)
def setup_teardown_module():
    """
    Remove log files at the end of this test module

    Before each test:
    - init logger
    - create directory to put generated files

    After:
    - remove all log files
    - remove directory with generated results
    """
    if not os.path.isdir(GENEPATH):
        print("setup")
        os.mkdir(GENEPATH)
    print("setup")

    yield
    shutil.rmtree(GENEPATH, ignore_errors=True)
    print("teardown")


def test_main_from_parse():
    """
    Test that when a tmp folder is given by user, tmp files are saved in it,
    and prokka files too.
    """
    list_file = os.path.join(TEST_DIR, "list_genomes-func-test-default.txt")
    tmpdir = os.path.join(GENEPATH, "tmp_funcGivenTmp")
    name = "ESCO"
    l90 = 100
    date = "0417"
    args = argparse.Namespace()
    args.list_file = list_file
    args.db_path = GEN_PATH
    args.res_path = GENEPATH
    args.name = name
    args.date = date
    args.l90 = l90
    args.nbcont = 999
    args.cutn = 0
    args.threads = 1
    args.force = False
    args.qc_only = False
    args.tmpdir = tmpdir
    args.prokkadir = None
    args.verbose = False
    args.quiet = False
    args.from_info = False
    args.prodigal_only = False
    args.small = False
    args.annotdir = False
    args.argv = ["annotate", "test_annote.py", "test_main_from_parse"]
    args.prodigal_only = False
    annot.main_from_parse(args)
    # Check that only 2 fna files (over 3) were created in tmp files. The 3rd
    # one is annotated from the original seq, nothing to merge
    assert len(glob.glob(os.path.join(tmpdir, '*.fna'))) == 2
    # Check that tmp files exist in the right folder
    assert os.path.isfile(os.path.join(tmpdir, "A_H738.fasta-all.fna"))
    assert os.path.isfile(os.path.join(tmpdir, "H299_H561.fasta-all.fna"))
    # Test that prokka folder is in the right directory
    assert os.path.isdir(os.path.join(tmpdir, "A_H738.fasta-all.fna-prokkaRes"))
    assert os.path.isdir(os.path.join(tmpdir, "B2_A3_5.fasta-changeName.fna-prokkaRes"))
    assert os.path.isdir(os.path.join(tmpdir, "H299_H561.fasta-all.fna-prokkaRes"))


def test_main_novalid_genome(capsys):
    """
    Test that when, in the list file, all genomes are wrong (do not correspond to
    filenames in the given dbpath), it closes the program with an error message.
    """
    list_file = os.path.join(TEST_DIR, "list_no_genome.txt")
    name = "ESCO"
    date = "0417"
    with pytest.raises(SystemExit):
        annot.main("cmd", list_file, GEN_PATH, GENEPATH, name, date)
    _, err = capsys.readouterr()
    assert ("We did not find any genome listed in test/data/annotate/test_files/"
            "list_no_genome.txt "
            "in the folder test/data/annotate/genomes. Please check your list to give valid "
            "genome names.") in err


def test_main_given_tmp_verbose3(capsys):
    """
    Test that when a tmp folder is given by user, tmp files are saved in it,
    and prokka files too.
    + check that, with verbose=3, warning and details are written to stdout

    Giving 4 genomes in list_files
    - for 1 genome, toto.fst does not exist, and will not be in the concatenated file
    - 2 concatenated files
    - 4 files to annotate
    - 4 prokkaRes
    - 1 genome with problems: no CDS found
    - 3 genomes in result dirs
    """
    list_file = os.path.join(TEST_DIR, "list_genomes-func-test-default.txt")
    tmpdir = os.path.join(GENEPATH, "tmp_funcGivenTmp")
    name = "ESCO"
    l90 = 10
    date = "0417"
    verbose = 3
    info_file = os.path.join(GENEPATH, "LSTINFO-list_genomes-func-test-default.lst")
    assert annot.main("cmd", list_file, GEN_PATH, GENEPATH, name, date, l90,
                      cutn=3, tmp_dir=tmpdir, verbose=verbose) == (info_file, 3)
    out, err = capsys.readouterr()
    # Check that warnings are written to stderr
    assert "WARNING" in err
    assert ("toto.fst genome file does not exist. Its file will be ignored when "
            "concatenating ['A_H738.fasta', 'genome1.fasta', 'toto.fst']") in err
    # Check that tmp files exist in the right folder
    # -> 2 fna files created (concatenations)
    # -> + 3 files created (split 5N)
    assert os.path.isfile(os.path.join(tmpdir, "A_H738.fasta-all.fna"))
    assert os.path.isfile(os.path.join(tmpdir, "H299_H561.fasta-all.fna"))
    assert len(glob.glob(os.path.join(tmpdir, '*.fna'))) == 6
    assert len(glob.glob(os.path.join(tmpdir, '*split3N.fna'))) == 4
    # Check that split contigs were renamed with unique ID at the begining of the header
    res_file = os.path.join(tmpdir, "A_H738.fasta-all.fna_prokka-split3N.fna")
    exp_file = os.path.join(EXP_DIR, "exp_A_H738.fasta-all.fna_prokka-split3N.fna")
    assert tutil.compare_order_content(exp_file, res_file)
    # Check that even for complete genome, contig was renamed with ID
    res_file = os.path.join(tmpdir, "complete_genome.fna_prokka-split3N.fna")
    exp_file = os.path.join(EXP_DIR, "exp_complete_genome.fna_prokka-split3N.fna")
    assert tutil.compare_order_content(exp_file, res_file)
    # Test that prokka folder is in the right directory
    # Only 1 genome annotated by  prokka (the 2 others do not have appropriate L90/nbcont)
    assert os.path.isdir(os.path.join(tmpdir, "A_H738.fasta-all.fna_prokka-split3N.fna-prokkaRes"))
    assert not os.path.isdir(os.path.join(tmpdir, "H299_H561.fasta-all.fna-prokkaRes"))


def test_main_all_discard_nbcont(capsys):
    """
    Test that when the genomes given in list file have high nbcontigs compared
    to given threshold, error message as there are no genome to annotate
    """
    list_file = os.path.join(TEST_DIR, "list_genomes-func-test-default.txt")
    name = "ESCO"
    nbcont = 0
    cutn = 0
    date = "0417"
    assert annot.main("cmd", list_file, GEN_PATH, GENEPATH, name, date, nbcont=nbcont,
                      cutn=cutn) == ("", 0)
    # check that there are the 2 concatenated genomes in tmppath.
    # The third genome is listfile is composed of only 1 file, so no need to concatenate, nor
    # to change the file as we do not use cutn
    assert os.path.isfile(os.path.join(GENEPATH, "tmp_files", "H299_H561.fasta-all.fna"))
    assert os.path.isfile(os.path.join(GENEPATH, "tmp_files", "A_H738.fasta-all.fna"))
    out, err = capsys.readouterr()
    assert 'No genome kept for annotation' in out


def test_main_qc():
    """
    Test that when only QC is run, it writes:
    - the list of all genomes with their characteristics
    - the list of genomes that would be discarded for annotation
    - the 2 png files
    """
    list_file = os.path.join(TEST_DIR, "list_genomes-func-test-default.txt")
    name = "ESCO"
    cutn = 0
    threads = 1
    l90 = 1
    date = "0417"
    force = False
    qc_only = True
    assert annot.main("cmd", list_file, GEN_PATH, GENEPATH, name, date, l90=l90,
                      cutn=cutn, qc_only=qc_only) == ("", 0)
    # Check files are here
    lstfile = os.path.join(GENEPATH, "ALL-GENOMES-info-list_genomes-func-test-default.lst")
    exp_lstfile = os.path.join(EXP_DIR, "exp_ALL-GENOMES-QC.lst")
    discardedfile = os.path.join(GENEPATH, "discarded-list_genomes-func-test-default.lst")
    exp_discarded = os.path.join(EXP_DIR, "exp_discarded_QC.lst")
    assert os.path.isfile(lstfile)
    assert os.path.isfile(discardedfile)
    assert os.path.isfile(os.path.join(GENEPATH,
                          "QC_L90-list_genomes-func-test-default.png"))
    assert os.path.isfile(os.path.join(GENEPATH,
                          "QC_nb-contigs-list_genomes-func-test-default.png"))
    # Check content of discarded genomes
    assert tutil.compare_file_content(lstfile, exp_lstfile)
    assert tutil.compare_file_content(discardedfile, exp_discarded)


def test_main_existresdirforce(capsys):
    """
    Test that, when the pipeline is run on an existing result directory, but force option is on,
    it removes the result folders and runs again.
    Result folders contain expected files, the ones put before are removed
    Giving 4 genomes
    - 4 are OK
    - trained on complete_genome_big.fna
    """
    list_file = os.path.join(TEST_DIR, "list_genomes-func-test-force.txt")
    # Create output directory with a prt file in Proteins folder
    protdir = os.path.join(GENEPATH, "Proteins")
    os.makedirs(protdir)
    open(os.path.join(protdir, "toto.prt"), "w").close()
    assert os.path.isfile(os.path.join(protdir, "toto.prt"))
    name = "ESCO"
    date = "0417"
    l90 = 5
    cutn = 3
    info_file = os.path.join(GENEPATH, "LSTINFO-list_genomes-func-test-force.lst")
    assert annot.main("cmd", list_file, GEN_PATH, GENEPATH, name, date, force=True, l90=l90,
                      prodigal_only=True, cutn = cutn) == (info_file, 4)
    out, err = capsys.readouterr()

    # Check that tmp files exist in the right folder
    # -> 2 fna files created (concatenations)
    # -> + 4 files created (split 3N)
    assert os.path.isfile(os.path.join(GENEPATH, "tmp_files", "A_H738.fasta-all.fna"))
    assert os.path.isfile(os.path.join(GENEPATH, "tmp_files", "H299_H561.fasta-all.fna"))
    assert len(glob.glob(os.path.join(GENEPATH, "tmp_files", '*.fna'))) == 6
    assert len(glob.glob(os.path.join(GENEPATH, "tmp_files", '*split3N.fna'))) == 4
    # Check that split contigs were renamed with unique ID at the begining of the header
    res_file = os.path.join(GENEPATH, "tmp_files", "A_H738.fasta-all.fna_prodigal-split3N.fna")
    exp_file = os.path.join(EXP_DIR, "exp_A_H738.fasta-all.fna_prokka-split3N.fna")
    assert tutil.compare_order_content(exp_file, res_file)
    # Check that even for complete genome, contig was renamed with ID
    res_file = os.path.join(GENEPATH, "tmp_files", "complete_genome_big.fna_prodigal-split3N.fna")
    exp_file = os.path.join(EXP_DIR, "exp_complete_genome_big.fna-split3N.fna")
    assert tutil.compare_order_content(exp_file, res_file)
    # Check that it trained on expected genome, and training file is ok
    trn_file = os.path.join(GENEPATH, "tmp_files", "complete_genome_big.fna"
                                                   "_prodigal-split3N.fna.trn")
    exp_trn_file = os.path.join(EXP_DIR, "exp_complete_genome_big.fna.trn")
    assert tutil.compare_files_bin(trn_file, exp_trn_file)
    # Check that tmp files exist in the right folder (result/tmp_files)
    assert os.path.isfile(os.path.join(GENEPATH, "tmp_files",
                                       "B2_A3_5.fasta-changeName.fna_prodigal-split3N.fna"))
    assert os.path.isfile(os.path.join(GENEPATH, "tmp_files",
                                       "H299_H561.fasta-all.fna_prodigal-split3N.fna"))
    assert os.path.isfile(os.path.join(GENEPATH, "tmp_files",
                                       "H299_H561.fasta-all.fna"))
    assert os.path.isfile(os.path.join(GENEPATH, "tmp_files",
                                       "A_H738.fasta-all.fna_prodigal-split3N.fna"))
    assert os.path.isfile(os.path.join(GENEPATH, "tmp_files",
                                       "A_H738.fasta-all.fna"))
    # Test all result folders are empty (in particular Proteins) as no genome is annotated
    assert os.path.isdir(protdir)
    assert len(os.listdir(protdir)) == 4
    assert not os.path.isfile(os.path.join(protdir, "toto.prt"))
    assert os.path.isfile(os.path.join(protdir, "ESCO.0417.00001.prt"))
    assert os.path.isfile(os.path.join(protdir, "ESCO.1015.00002.prt"))
    assert os.path.isfile(os.path.join(protdir, "ESCO.1015.00003.prt"))
    assert os.path.isfile(os.path.join(protdir, "ESCO.1116.00004.prt"))


def test_run_exist_resdir(caplog):
    """
    Test that when the pipeline is called, with a given resdir which already contains
    results, the program ends, with an error message.
    """
    # Create output directory with a lst file in LSTINFO
    os.makedirs(os.path.join(GENEPATH, "Proteins"))
    open(os.path.join(GENEPATH, "Proteins", "toto.prt"), "w").close()
    with pytest.raises(SystemExit):
        annot.main("cmd", "list_file.lst", "path/db", GENEPATH, "toto", "0123")
    assert ("ERROR: Your output directory already has .prt files in the "
            "Proteins folder. Provide another result directory, or remove the "
            "files in this one.") in caplog.text
    # File was not removed
    assert os.path.isfile(os.path.join(GENEPATH, "Proteins", "toto.prt"))


def test_main_onexistingprokkadir(capsys):
    """
    Test that, when the pipeline is run with a given prokka dir, where prokka results already
    exist, and are ok, all runs well, no re-annotation, just format


    main function arguments:
    cmd, list_file, db_path, res_dir, name, date, l90=100, nbcont=999, cutn=5,
    threads=1, force=False, qc_only=False, from_info=None, tmp_dir=None, res_annot_dir=None,
    verbose=0, quiet=False, prodigal_only=False):

    """
    list_file = os.path.join(TEST_DIR, "list_genomes-func-test-exist_dir.txt")
    name = "ESCO"
    date = "0417"
    lstout = os.path.join(GENEPATH, "LSTINFO-list_genomes-func-test-exist_dir.lst")
    lstexp = os.path.join(EXP_DIR, "exp_LSTINFO-func-annot_exists-prokkadir.lst")
    assert annot.main("cmd", list_file, GEN_PATH, GENEPATH, name, date, cutn=0,
                      res_annot_dir=EXP_DIR, verbose=3) == (lstout, 2)
    out, err = capsys.readouterr()
    # Check that tmp files folder is empty (prokka res are somewhere else)
    assert len(os.listdir(os.path.join(GENEPATH, "tmp_files"))) == 0
    # Test that result files are in result dir
    assert os.path.isfile(lstout)
    assert tutil.compare_order_content(lstout, lstexp)
    logfile = os.path.join(GENEPATH,
                           "PanACoTA-annotate_list_genomes-func-test-exist_dir.log.details")
    with open(logfile, "r") as lc:
        log_content = lc.readlines()
    assert ("Prokka results folder "
            "test/data/annotate/exp_files/B2_A3_5.fasta-changeName.fna-prokkaRes "
            "already exists") in " ".join(log_content)
    assert ("Prokka did not run again, formatting step used already generated results of Prokka "
            "in test/data/annotate/exp_files/H299_H561.fasta-prokkaRes. "
            "If you want to re-run prokka, first remove this result folder, or use '-F' or "
            "'--force' option if you want to rerun prokka "
            "for all genomes.") in ' '.join(log_content)


def test_main_onexistingprodigaldir(capsys):
    """
    Test that, when the pipeline is run with a given prodigal dir, where prodigal results already
    exist, and are ok, all runs well, no re-annotation, just format

    - trains
    - no re-annotation
    - format

    main function arguments:
    cmd, list_file, db_path, res_dir, name, date, l90=100, nbcont=999, cutn=5,
         threads=1, force=False, qc_only=False, from_info=None, tmp_dir=None, res_annot_dir=None,
         verbose=0, quiet=False, prodigal_only=False

    """
    # FOLDER with all results
    # Create result folder, with existing prodigal folders (which are OK)
    res_folder = os.path.join(GENEPATH, "results-prodigal")
    os.makedirs(res_folder)
    # copy prodigalRes folders
    B2_A3_5_folder = os.path.join(EXP_DIR, "B2_A3_5.fasta-changeName.fna-prodigalRes")
    H299_folder = os.path.join(EXP_DIR, "H299_H561.fasta-prodigalRes")
    res_B2_A3_5_folder = os.path.join(res_folder, "B2_A3_5.fasta-changeName.fna-prodigalRes")
    res_H299_folder = os.path.join(res_folder, "H299_H561.fasta-prodigalRes")
    shutil.copytree(B2_A3_5_folder, res_B2_A3_5_folder)
    shutil.copytree(H299_folder, res_H299_folder)

    list_file = os.path.join(TEST_DIR, "list_genomes-func-test-exist_dir.txt")
    name = "ESCO"
    date = "0417"
    lstout = os.path.join(GENEPATH, "LSTINFO-list_genomes-func-test-exist_dir.lst")
    lstexp = os.path.join(EXP_DIR, "exp_LSTINFO-func-annot_exists-prokkadir.lst")
    assert annot.main("cmd", list_file, GEN_PATH, GENEPATH, name, date, cutn=0,
                      res_annot_dir=res_folder, verbose=3, prodigal_only=True) == (lstout, 2)
    out, err = capsys.readouterr()
    # Check that tmp files folder is empty (prokka res are somewhere else)
    assert len(os.listdir(os.path.join(GENEPATH, "tmp_files"))) == 0
    # Test that result files are in result dir
    assert os.path.isfile(lstout)
    assert tutil.compare_order_content(lstout, lstexp)
    logfile = os.path.join(GENEPATH,
                           "PanACoTA-annotate_list_genomes-func-test-exist_dir.log.details")
    with open(logfile, "r") as lc:
        log_content = lc.readlines()
    assert ("Prodigal will train using "
            "test/data/annotate/genomes/H299_H561.fasta") in " ".join(log_content)
    assert ("prodigal command: prodigal -i test/data/annotate/genomes/H299_H561.fasta "
            "-t test/data/annotate/generated_by_func-tests/results-prodigal/H299_H561.fasta.trn")
    assert("Error while trying to train prodigal on H299_H561.fasta. "
           "See test/data/annotate/generated_by_func-tests/results-prodigal/"
           "H299_H561.fasta.trn-prodigal-train.log.err") in " ".join(log_content)
    assert ("Prodigal results folder test/data/annotate/generated_by_func-tests/"
            "results-prodigal/H299_H561.fasta-prodigalRes "
            "already exists") in " ".join(log_content)
    assert ("Prodigal did not run again. Formatting step will use already generated results of "
            "Prodigal in test/data/annotate/generated_by_func-tests/results-prodigal/"
            "H299_H561.fasta-prodigalRes. "
            "If you want to re-run Prodigal, first remove this result folder, or use '-F' or "
            "'--force' option.") in ' '.join(log_content)
    assert "Formatting all genomes" in " ".join(log_content)
    assert "Annotation step done" in " ".join(log_content)


def test_main_onexistingprodigaldir_train_exists(capsys):
    """
    Test that, when the pipeline is run with a given prodigal dir, where prodigal results already
    exist, and are ok, all runs well, no re-annotation, just format

    - no train
    - no reannote
    - format

    2 genomes in list file: B2_A3_5.fasta-changeName.fna and H299_H561.fasta
    """
    # FOLDER with all results
    # Create result folder, with existing prodigal folders (which are OK)
    res_folder = os.path.join(GENEPATH, "results-prodigal")
    os.makedirs(res_folder)
    # copy prodigalRes folders
    B2_A3_5_folder = os.path.join(EXP_DIR, "B2_A3_5.fasta-changeName.fna-prodigalRes")
    H299_folder = os.path.join(EXP_DIR, "H299_H561.fasta-prodigalRes")
    res_B2_A3_5_folder = os.path.join(res_folder, "B2_A3_5.fasta-changeName.fna-prodigalRes")
    res_H299_folder = os.path.join(res_folder, "H299_H561.fasta-prodigalRes")
    shutil.copytree(B2_A3_5_folder, res_B2_A3_5_folder)
    shutil.copytree(H299_folder, res_H299_folder)
    # Add a training file in result folder
    trn_file = os.path.join(res_folder, "H299_H561.fasta.trn")
    open(trn_file, "w").close()

    # Function arguments
    list_file = os.path.join(TEST_DIR, "list_genomes-func-test-exist_dir.txt")
    name = "ESCO"
    date = "0417"
    lstout = os.path.join(GENEPATH, "LSTINFO-list_genomes-func-test-exist_dir.lst")
    lstexp = os.path.join(EXP_DIR, "exp_LSTINFO-func-annot_exists-prokkadir.lst")
    assert annot.main("cmd", list_file, GEN_PATH, GENEPATH, name, date, cutn=0,
                      res_annot_dir=res_folder, verbose=3, prodigal_only=True) == (lstout, 2)
    out, err = capsys.readouterr()
    # Check that tmp files folder is empty (prokka res are somewhere else)
    assert len(os.listdir(os.path.join(GENEPATH, "tmp_files"))) == 0
    # Test that result files are in result dir
    assert os.path.isfile(lstout)
    assert tutil.compare_order_content(lstout, lstexp)
    logfile = os.path.join(GENEPATH,
                           "PanACoTA-annotate_list_genomes-func-test-exist_dir.log.details")
    with open(logfile, "r") as lc:
        log_content = lc.readlines()
    assert ("A training file already exists (test/data/annotate/generated_by_func-tests/"
            "results-prodigal/H299_H561.fasta.trn). It will be used to annotate "
            "all genomes.") in " ".join(log_content)
    assert ("Prodigal results folder test/data/annotate/generated_by_func-tests/"
            "results-prodigal/H299_H561.fasta-prodigalRes "
            "already exists") in " ".join(log_content)
    assert ("Prodigal results folder test/data/annotate/generated_by_func-tests/"
            "results-prodigal/B2_A3_5.fasta-changeName.fna-prodigalRes "
            "already exists") in " ".join(log_content)
    assert ("Prodigal did not run again. Formatting step will use already generated results of "
            "Prodigal in test/data/annotate/generated_by_func-tests/results-prodigal/"
            "H299_H561.fasta-prodigalRes. "
            "If you want to re-run Prodigal, first remove this result folder, or use '-F' or "
            "'--force' option.") in ' '.join(log_content)
    assert "Formatting all genomes" in " ".join(log_content)
    assert "Annotation step done" in " ".join(log_content)


def test_main_prodigal_train_empty(capsys):
    """
    Test that, when the pipeline is run with a given prodigal dir, where prodigal results do
    not exist, and the given trn file is empty
    -> error, with prodigal command

    - no train
    - try reannote but fails -> exits

    2 genomes in list file: B2_A3_5.fasta-changeName.fna and H299_H561.fasta
    """
    # FOLDER with all results
    # Create result folder, with empty trn file
    res_folder = os.path.join(GENEPATH, "results-prodigal")
    os.makedirs(res_folder)
    trn_file = os.path.join(res_folder, "H299_H561.fasta.trn")
    open(trn_file, "w").close()

    # Function arguments
    list_file = os.path.join(TEST_DIR, "list_genomes-func-test-exist_dir.txt")
    name = "ESCO"
    date = "0417"
    lstout = os.path.join(GENEPATH, "LSTINFO-list_genomes-func-test-exist_dir.lst")
    lstexp = os.path.join(EXP_DIR, "exp_LSTINFO-func-annot_exists-prokkadir.lst")
    with pytest.raises(SystemExit):
        annot.main("cmd", list_file, GEN_PATH, GENEPATH, name, date, cutn=0,
                   res_annot_dir=res_folder, verbose=3, prodigal_only=True)
    out, err = capsys.readouterr()
    # Test that result files are in result dir
    assert os.path.isfile(lstout)
    assert tutil.compare_order_content(lstout, lstexp)
    logfile = os.path.join(GENEPATH,
                           "PanACoTA-annotate_list_genomes-func-test-exist_dir.log.details")
    # Check logs
    with open(logfile, "r") as lc:
        log_content = lc.readlines()
    assert ("A training file already exists (test/data/annotate/generated_by_func-tests/"
            "results-prodigal/H299_H561.fasta.trn). It will be used to annotate "
            "all genomes.") in " ".join(log_content)
    assert ("Start annotating ESCO.1116.00002 (from test/data/annotate/genomes/"
            "B2_A3_5.fasta-changeName.fna sequence) with Prodigal") in " ".join(log_content)
    assert ("Start annotating ESCO.1015.00001 (from test/data/annotate/genomes/"
            "H299_H561.fasta sequence) with Prodigal") in " ".join(log_content)
    assert ("Prodigal command: "
            "prodigal -i test/data/annotate/genomes/B2_A3_5.fasta-changeName.fna "
            "-d test/data/annotate/generated_by_func-tests/results-prodigal/"
            "B2_A3_5.fasta-changeName.fna-prodigalRes/ESCO.1116.00002.ffn "
            "-a test/data/annotate/generated_by_func-tests/results-prodigal/"
            "B2_A3_5.fasta-changeName.fna-prodigalRes/ESCO.1116.00002.faa "
            "-f gff -o test/data/annotate/generated_by_func-tests/results-prodigal/"
            "B2_A3_5.fasta-changeName.fna-prodigalRes/ESCO.1116.00002.gff "
            "-t test/data/annotate/generated_by_func-tests/results-prodigal/H299_H561.fasta.trn "
            "-q") in " ".join(log_content)
    assert ("Prodigal command: "
            "prodigal -i test/data/annotate/genomes/H299_H561.fasta "
            "-d test/data/annotate/generated_by_func-tests/results-prodigal/"
            "H299_H561.fasta-prodigalRes/ESCO.1015.00001.ffn "
            "-a test/data/annotate/generated_by_func-tests/results-prodigal/"
            "H299_H561.fasta-prodigalRes/ESCO.1015.00001.faa "
            "-f gff -o test/data/annotate/generated_by_func-tests/results-prodigal/"
            "H299_H561.fasta-prodigalRes/ESCO.1015.00001.gff "
            "-t test/data/annotate/generated_by_func-tests/results-prodigal/H299_H561.fasta.trn "
            "-q") in " ".join(log_content)
    assert ("Error while trying to run prodigal. See test/data/annotate/generated_by_func-tests/"
            "results-prodigal/H299_H561.fasta-prodigal.log.err.") in " ".join(log_content)
    assert ("Error while trying to run prodigal. See test/data/annotate/generated_by_func-tests/"
            "results-prodigal/"
            "B2_A3_5.fasta-changeName.fna-prodigal.log.err") in " ".join(log_content)
    assert ("No genome was correctly annotated, no need to format them.") in " ".join(log_content)


def test_main_prodigal_train_ok(capsys):
    """
    Test that, when the pipeline is run with a given prodigal dir, where prodigal train exists and is ok:

    - no train
    - reannotate
    - format

    2 genomes in list file: B2_A3_5.fasta-changeName.fna and H299_H561.fasta
    """
    # FOLDER with all results
    # Create result folder, with existing prodigal folders (which are OK)
    res_folder = os.path.join(GENEPATH, "results-prodigal")
    os.makedirs(res_folder)
    # Add a valid training file in result folder
    orig_trn_file = os.path.join(EXP_DIR, "exp_complete_genome_big.fna.trn")
    trn_file = os.path.join(res_folder, "H299_H561.fasta.trn")
    shutil.copyfile(orig_trn_file, trn_file)

    # Function arguments
    list_file = os.path.join(TEST_DIR, "list_genomes-func-test-exist_dir.txt")
    name = "ESCO"
    date = "0417"
    lstout = os.path.join(GENEPATH, "LSTINFO-list_genomes-func-test-exist_dir.lst")
    lstexp = os.path.join(EXP_DIR, "exp_LSTINFO-func-annot_exists-prokkadir.lst")
    assert annot.main("cmd", list_file, GEN_PATH, GENEPATH, name, date, cutn=0,
                      res_annot_dir=res_folder, verbose=3, prodigal_only=True) == (lstout, 2)
    out, err = capsys.readouterr()
    # Check that tmp files folder is empty (prokka res are somewhere else)
    assert len(os.listdir(os.path.join(GENEPATH, "tmp_files"))) == 0
    # Test that result files are in result dir
    assert os.path.isfile(lstout)
    assert tutil.compare_order_content(lstout, lstexp)
    logfile = os.path.join(GENEPATH,
                           "PanACoTA-annotate_list_genomes-func-test-exist_dir.log.details")
    with open(logfile, "r") as lc:
        log_content = lc.readlines()
    assert ("A training file already exists (test/data/annotate/generated_by_func-tests/"
            "results-prodigal/H299_H561.fasta.trn). It will be used to annotate "
            "all genomes.") in " ".join(log_content)
    assert ("Start annotating ESCO.1116.00002 (from test/data/annotate/genomes/"
            "B2_A3_5.fasta-changeName.fna sequence) with Prodigal") in " ".join(log_content)
    assert ("Start annotating ESCO.1015.00001 (from test/data/annotate/genomes/"
            "H299_H561.fasta sequence) with Prodigal") in " ".join(log_content)
    assert ("Prodigal command: "
            "prodigal -i test/data/annotate/genomes/B2_A3_5.fasta-changeName.fna "
            "-d test/data/annotate/generated_by_func-tests/results-prodigal/"
            "B2_A3_5.fasta-changeName.fna-prodigalRes/ESCO.1116.00002.ffn "
            "-a test/data/annotate/generated_by_func-tests/results-prodigal/"
            "B2_A3_5.fasta-changeName.fna-prodigalRes/ESCO.1116.00002.faa "
            "-f gff -o test/data/annotate/generated_by_func-tests/results-prodigal/"
            "B2_A3_5.fasta-changeName.fna-prodigalRes/ESCO.1116.00002.gff "
            "-t test/data/annotate/generated_by_func-tests/results-prodigal/H299_H561.fasta.trn "
            "-q") in " ".join(log_content)
    assert ("Prodigal command: "
            "prodigal -i test/data/annotate/genomes/H299_H561.fasta "
            "-d test/data/annotate/generated_by_func-tests/results-prodigal/"
            "H299_H561.fasta-prodigalRes/ESCO.1015.00001.ffn "
            "-a test/data/annotate/generated_by_func-tests/results-prodigal/"
            "H299_H561.fasta-prodigalRes/ESCO.1015.00001.faa "
            "-f gff -o test/data/annotate/generated_by_func-tests/results-prodigal/"
            "H299_H561.fasta-prodigalRes/ESCO.1015.00001.gff "
            "-t test/data/annotate/generated_by_func-tests/results-prodigal/H299_H561.fasta.trn "
            "-q") in " ".join(log_content)
    assert ("End annotating ESCO.1015.00001 (from test/data/annotate/genomes/"
            "H299_H561.fasta)") in " ".join(log_content)
    assert "Formatting all genomes" in " ".join(log_content)
    assert "Annotation step done" in " ".join(log_content)


def test_main_prodigal_ok(capsys):
    """
    Test that, when the pipeline is run with a given empty prodigal dir, it does everything

    - train
    - reannotate
    - format

    2 genomes in list file: B2_A3_5.fasta-changeName.fna and H299_H561.fasta
    """
    # FOLDER with all results
    # Create result folder, with existing prodigal folders (which are OK)
    res_folder = os.path.join(GENEPATH, "results-prodigal")
    os.makedirs(res_folder)

    # Function arguments
    list_file = os.path.join(GENEPATH, "list_genomes.txt")
    with open(list_file, "w") as lf:
        lf.write("A_H738.fasta \n")
        lf.write("H299_H561.fasta::TOTO")
    name = "ESCO"
    date = "0417"
    lstout = os.path.join(GENEPATH, "LSTINFO-list_genomes.lst")
    lstexp = os.path.join(EXP_DIR, "exp_LSTINFO-func-annot_exists-prokkadir.lst")
    assert annot.main("cmd", list_file, GEN_PATH, GENEPATH, name, date, cutn=0,
                      res_annot_dir=res_folder, verbose=3, prodigal_only=True) == (lstout, 2)
    out, err = capsys.readouterr()
    # Test that result files are in result dir
    assert os.path.isfile(lstout)
    logfile = os.path.join(GENEPATH,
                           "PanACoTA-annotate_list_genomes.log.details")
    with open(logfile, "r") as lc:
        log_content = lc.readlines()
    assert ("Prodigal will train using "
            "test/data/annotate/genomes/A_H738.fasta") in " ".join(log_content)
    assert ("prodigal command: prodigal -i test/data/annotate/genomes/A_H738.fasta "
            "-t test/data/annotate/generated_by_func-tests/results-prodigal/"
            "A_H738.fasta.trn") in " ".join(log_content)
    assert ("Start annotating ESCO.0417.00001 (from test/data/annotate/genomes/"
            "A_H738.fasta sequence) with Prodigal") in " ".join(log_content)
    assert ("Start annotating TOTO.0417.00001 (from test/data/annotate/genomes/"
            "H299_H561.fasta sequence) with Prodigal") in " ".join(log_content)
    assert ("Prodigal command: "
            "prodigal -i test/data/annotate/genomes/A_H738.fasta "
            "-d test/data/annotate/generated_by_func-tests/results-prodigal/"
            "A_H738.fasta-prodigalRes/ESCO.0417.00001.ffn "
            "-a test/data/annotate/generated_by_func-tests/results-prodigal/"
            "A_H738.fasta-prodigalRes/ESCO.0417.00001.faa "
            "-f gff -o test/data/annotate/generated_by_func-tests/results-prodigal/"
            "A_H738.fasta-prodigalRes/ESCO.0417.00001.gff "
            "-t test/data/annotate/generated_by_func-tests/results-prodigal/A_H738.fasta.trn "
            "-q") in " ".join(log_content)
    assert ("Prodigal command: "
            "prodigal -i test/data/annotate/genomes/H299_H561.fasta "
            "-d test/data/annotate/generated_by_func-tests/results-prodigal/"
            "H299_H561.fasta-prodigalRes/TOTO.0417.00001.ffn "
            "-a test/data/annotate/generated_by_func-tests/results-prodigal/"
            "H299_H561.fasta-prodigalRes/TOTO.0417.00001.faa "
            "-f gff -o test/data/annotate/generated_by_func-tests/results-prodigal/"
            "H299_H561.fasta-prodigalRes/TOTO.0417.00001.gff "
            "-t test/data/annotate/generated_by_func-tests/results-prodigal/A_H738.fasta.trn "
            "-q") in " ".join(log_content)
    assert ("End annotating ESCO.0417.00001 (from test/data/annotate/genomes/"
            "A_H738.fasta)") in " ".join(log_content)
    assert ("End annotating TOTO.0417.00001 (from test/data/annotate/genomes/"
            "H299_H561.fasta)") in " ".join(log_content)
    assert "Formatting all genomes" in " ".join(log_content)
    assert "Annotation step done" in " ".join(log_content)


def test_main_prodigal_small_ok(capsys):
    """
    Test that, when the pipeline is run with a given prodigal dir, and --small option, it does:

    - no train
    - reannotate
    - format

    2 genomes in list file: B2_A3_5.fasta-changeName.fna and H299_H561.fasta
    """
    # FOLDER with all results
    # Create result folder, with existing prodigal folders (which are OK)
    res_folder = os.path.join(GENEPATH, "results-prodigal")
    os.makedirs(res_folder)

    # Function arguments
    list_file = os.path.join(GENEPATH, "list_genomes_prodigal_small.txt")
    with open(list_file, "w") as lf:
        lf.write("A_H738.fasta \n")
        lf.write("H299_H561.fasta::TOTO")
    name = "ESCO"
    date = "0417"
    lstout = os.path.join(GENEPATH, "LSTINFO-list_genomes_prodigal_small.lst")
    # lstexp = os.path.join(EXP_DIR, "exp_LSTINFO-func-annot_exists-prokkadir.lst")
    assert annot.main("cmd", list_file, GEN_PATH, GENEPATH, name, date, cutn=0,
                      res_annot_dir=res_folder, verbose=3, prodigal_only=True, small=True) == (lstout, 2)
    out, err = capsys.readouterr()
    # Test that result files are in result dir
    assert os.path.isfile(lstout)
    logfile = os.path.join(GENEPATH,
                           "PanACoTA-annotate_list_genomes_prodigal_small.log.details")
    with open(logfile, "r") as lc:
        log_content = lc.readlines()
    assert ("Start annotating ESCO.0417.00001 (from test/data/annotate/genomes/"
            "A_H738.fasta sequence) with Prodigal") in " ".join(log_content)
    assert ("Start annotating TOTO.0417.00001 (from test/data/annotate/genomes/"
            "H299_H561.fasta sequence) with Prodigal") in " ".join(log_content)
    assert ("Prodigal command: "
            "prodigal -i test/data/annotate/genomes/A_H738.fasta "
            "-d test/data/annotate/generated_by_func-tests/results-prodigal/"
            "A_H738.fasta-prodigalRes/ESCO.0417.00001.ffn "
            "-a test/data/annotate/generated_by_func-tests/results-prodigal/"
            "A_H738.fasta-prodigalRes/ESCO.0417.00001.faa "
            "-f gff -o test/data/annotate/generated_by_func-tests/results-prodigal/"
            "A_H738.fasta-prodigalRes/ESCO.0417.00001.gff "
            "-p meta -q") in " ".join(log_content)
    assert ("Prodigal command: "
            "prodigal -i test/data/annotate/genomes/H299_H561.fasta "
            "-d test/data/annotate/generated_by_func-tests/results-prodigal/"
            "H299_H561.fasta-prodigalRes/TOTO.0417.00001.ffn "
            "-a test/data/annotate/generated_by_func-tests/results-prodigal/"
            "H299_H561.fasta-prodigalRes/TOTO.0417.00001.faa "
            "-f gff -o test/data/annotate/generated_by_func-tests/results-prodigal/"
            "H299_H561.fasta-prodigalRes/TOTO.0417.00001.gff "
            "-p meta -q") in " ".join(log_content)
    assert ("End annotating ESCO.0417.00001 (from test/data/annotate/genomes/"
            "A_H738.fasta)") in " ".join(log_content)
    assert ("End annotating TOTO.0417.00001 (from test/data/annotate/genomes/"
            "H299_H561.fasta)") in " ".join(log_content)
    assert "Formatting all genomes" in " ".join(log_content)
    assert "Annotation step done" in " ".join(log_content)


def test_main_existing_prokkadir_errorannot():
    """
    Test that, when the pipeline is run with a given prokka dir, where prokka results have
    problems (no tbl file and no gff file), it returns an error message and the genome with
    problems is in skipped.
    """
    list_file = os.path.join(TEST_DIR, "list_genomes-func-test-exist_dir.txt")

    # Create directory with all files needed for the test
    genome_path_used = os.path.join(GENEPATH, "genomes")
    os.makedirs(genome_path_used)
    ori_genome1 = os.path.join(GEN_PATH, "B2_A3_5.fasta-changeName.fna")
    ori_prok_g1 = os.path.join(EXP_DIR, "B2_A3_5.fasta-changeName.fna-prokkaRes")
    used_genome1 = os.path.join(genome_path_used, "B2_A3_5.fasta-changeName.fna")
    used_prok_g1 = used_genome1 + "-prokkaRes"
    # Copy original fasta file to genepath/genomes
    shutil.copyfile(ori_genome1, used_genome1)
    # Copy prokka results to genepath/genomes/gname-prokkaRes
    shutil.copytree(ori_prok_g1, used_prok_g1)
    # Same think for 2nd genome
    ori_genome2 = os.path.join(GEN_PATH, "H299_H561.fasta")
    ori_prok_g2 = os.path.join(EXP_DIR, "H299_H561.fasta-prokkaRes")
    used_genome2 = os.path.join(genome_path_used, "H299_H561.fasta")
    used_prok_g2 = used_genome2 + "-prokkaRes"
    shutil.copyfile(ori_genome2, used_genome2)
    shutil.copytree(ori_prok_g2, used_prok_g2)

    # Remove tbl file for genome1, so that check_prokka returns an error
    os.remove(os.path.join(used_prok_g1, "ESCO.1116.00002.tbl"))

    # # Run annotation
    name = "ESCO"
    date = "0417"
    info_file = os.path.join(GENEPATH, "LSTINFO-list_genomes-func-test-exist_dir.lst")
    assert annot.main("cmd", list_file, genome_path_used, GENEPATH, name, date, cutn=0,
                      res_annot_dir=genome_path_used) == (info_file, 1)

    # Check that only 1 genome was formated (the other one had problems with prokka)
    prot_dir = os.path.join(GENEPATH, "Proteins")
    assert len(os.listdir(prot_dir)) == 1
    rep_dir = os.path.join(GENEPATH, "Replicons")
    assert len(os.listdir(rep_dir)) == 1

    # Check that the genome formated was not re-annotated
    logfile = os.path.join(GENEPATH,
                           "PanACoTA-annotate_list_genomes-func-test-exist_dir.log.details")
    with open(logfile, "r") as lf:
        log_content = lf.readlines()
    assert ("Prokka results folder "
            "test/data/annotate/generated_by_func-tests/genomes/H299_H561.fasta-prokkaRes "
            "already exists") in " ".join(log_content)
    # Check that genome not formated because error in prokka res
    assert ("ESCO.1116.00002 B2_A3_5.fasta-changeName.fna: no .tbl file") in " ".join(log_content)
    assert ("Problems in the files contained in your already existing output dir "
            "(test/data/annotate/generated_by_func-tests/genomes/"
            "B2_A3_5.fasta-changeName.fna-prokkaRes). Please check it, "
            "or remove it to re-annotate.") in ' '.join(log_content)


def test_main_existing_prodigaldir_errorannot(capsys):
    """
    Test that, when the pipeline is run with a given prodigal dir, where prodigal results have
    problems for both genomes, it returns an error message and the genomes with
    problems are in skipped. Error message with no genome to format.
    """
    list_file = os.path.join(TEST_DIR, "list_genomes-func-test-exist_dir.txt")

    # Create directory with all files needed for the test
    genome_path_used = os.path.join(GENEPATH, "genomes")
    os.makedirs(genome_path_used)
    ori_genome1 = os.path.join(GEN_PATH, "B2_A3_5.fasta-changeName.fna")
    ori_prok_g1 = os.path.join(EXP_DIR, "B2_A3_5.fasta-changeName.fna-prodigalRes")
    used_genome1 = os.path.join(genome_path_used, "B2_A3_5.fasta-changeName.fna")
    used_prok_g1 = used_genome1 + "-prodigalRes"
    # Copy original fasta file to genepath/genomes
    shutil.copyfile(ori_genome1, used_genome1)
    # Copy prokka results to genepath/genomes/gname-prokkaRes
    shutil.copytree(ori_prok_g1, used_prok_g1)
    # Same thing for 2nd genome
    ori_genome2 = os.path.join(GEN_PATH, "H299_H561.fasta")
    ori_prok_g2 = os.path.join(EXP_DIR, "H299_H561.fasta-prodigalRes")
    used_genome2 = os.path.join(genome_path_used, "H299_H561.fasta")
    used_prok_g2 = used_genome2 + "-prodigalRes"
    shutil.copyfile(ori_genome2, used_genome2)
    shutil.copytree(ori_prok_g2, used_prok_g2)

    # Remove faa file for genome1, so that check_prodigal returns an error
    os.remove(os.path.join(used_prok_g1, "ESCO.1116.00002.faa"))
    # Remove gff file for genome1, so that check_prodigal returns an error
    os.remove(os.path.join(used_prok_g2, "ESCO.1015.00001.gff"))

    # Run annotation
    name = "ESCO"
    date = "0417"
    with pytest.raises(SystemExit):
        annot.main("cmd", list_file, genome_path_used, GENEPATH, name, date, cutn=0,
                   res_annot_dir=genome_path_used, prodigal_only=True, verbose=15)

    # Check that Replicons & co folders are not created
    prot_dir = os.path.join(GENEPATH, "Proteins")
    assert not os.path.isdir(prot_dir)
    rep_dir = os.path.join(GENEPATH, "Replicons")
    assert not os.path.isdir(rep_dir)

    # Check that not formatted because exists + error
    logfile = os.path.join(GENEPATH,
                           "PanACoTA-annotate_list_genomes-func-test-exist_dir.log.details")
    with open(logfile, "r") as lc:
        log_content = lc.readlines()
    assert ("Error: No genome was correctly annotated, "
            "no need to format them") in ' '.join(log_content)
    assert ("Prodigal results folder test/data/annotate/generated_by_func-tests/genomes/"
            "H299_H561.fasta-prodigalRes already exists.") in ' '.join(log_content)
    assert ("ESCO.1116.00002 B2_A3_5.fasta-changeName.fna: "
            "no or several .faa file(s)") in ' '.join(log_content)
    assert ("ESCO.1015.00001 H299_H561.fasta: "
            "no or several .gff file(s)") in ' '.join(log_content)


def test_main_existing_prokkadir_errorformat():
    """
    Test that, when the pipeline is run with a given prokka dir, where prokka results are ok
    (all expected files), but have problems inside (wrong header format), it returns an error
    message and the genome with problems is in skipped_format.
    """
    list_file = os.path.join(TEST_DIR, "list_genomes-func-test-exist_dir.txt")

    # Create directory with all files needed for the test
    genome_path_used = os.path.join(GENEPATH, "genomes")
    os.makedirs(genome_path_used)
    ori_genome1 = os.path.join(GEN_PATH, "B2_A3_5.fasta-changeName.fna")
    # orig prokka dir has a tbl file with wrong format
    ori_prok_g1 = os.path.join(EXP_DIR, "B2_A3_5.fasta-changeName.fna-short-contig.fna-prokkaRes")
    used_genome1 = os.path.join(genome_path_used, "B2_A3_5.fasta-changeName.fna")
    used_prok_g1 = used_genome1 + "-prokkaRes"
    # Copy original fasta file to genepath/genomes
    shutil.copyfile(ori_genome1, used_genome1)
    # Copy prokka results to genepath/genomes/gname-prokkaRes
    shutil.copytree(ori_prok_g1, used_prok_g1)
    # and add .fna file to prokka-dir
    used_fna1 = os.path.join(used_prok_g1, "B2_A3_5.fasta-changeName.fna")
    shutil.copyfile(ori_genome1, used_fna1)

    # Same thing for 2nd genome
    ori_genome2 = os.path.join(GEN_PATH, "H299_H561.fasta")
    ori_prok_g2 = os.path.join(EXP_DIR, "H299_H561.fasta-prokkaRes")
    used_genome2 = os.path.join(genome_path_used, "H299_H561.fasta")
    used_prok_g2 = used_genome2 + "-prokkaRes"
    # Copy original fasta file to tmp resdir
    shutil.copyfile(ori_genome2, used_genome2)
    # Copy folder with prokka result files to genepath result path
    shutil.copytree(ori_prok_g2, used_prok_g2)
    # and add .fna file to prokka-dir
    used_fna = os.path.join(used_prok_g2, "H299_H561.fasta")
    shutil.copyfile(ori_genome2, used_fna)

    # Run annotation
    name = "ESCO"
    date = "0417"
    info_file = os.path.join(GENEPATH, "LSTINFO-list_genomes-func-test-exist_dir.lst")
    assert annot.main("cmd", list_file, genome_path_used, GENEPATH, name, date, cutn=0,
               res_annot_dir=genome_path_used) == (info_file, 1)

    # Check that genome 1 is not formatted, while no error with prokka
    logfile = os.path.join(GENEPATH,
                           "PanACoTA-annotate_list_genomes-func-test-exist_dir.log.details")
    with open(logfile, "r") as lc:
        log_content = lc.readlines()
    assert ("Prokka results folder test/data/annotate/generated_by_func-tests/genomes/"
            "B2_A3_5.fasta-changeName.fna-prokkaRes already exists.") in ' '.join(log_content)
    assert ("Prokka did not run again, formatting step used already generated "
            "results of Prokka in test/data/annotate/generated_by_func-tests/genomes/"
            "B2_A3_5.fasta-changeName.fna-prokkaRes.") in ' '.join(log_content)
    # Error while trying to format:
    assert ("'changesHead.0417.00010.0005' found in "
            "test/data/annotate/generated_by_func-tests/genomes/"
            "B2_A3_5.fasta-changeName.fna-prokkaRes/test.0417.00002.tbl "
            "does not exist in test/data/annotate/generated_by_func-tests/genomes/"
            "B2_A3_5.fasta-changeName.fna") in ' '.join(log_content)
    assert ("Problems while generating LSTINFO file for ESCO.1116.00002") in ' '.join(log_content)


def test_main_frominfo(capsys):
    """
    test that it runs well when giving an info file instead of list file + db etc.
    It does not re-calculate L90 and nbcont
    """
    listfile = None
    dbpath = None
    name = "TOTO"
    date = "1205"
    infofile = os.path.join(TEST_DIR, "lstinfo.lst")
    out_infofile = os.path.join(GENEPATH, "LSTINFO-lstinfo.lst")
    assert annot.main("cmd", listfile, dbpath, GENEPATH, name, date, from_info=infofile,
                      prodigal_only=True) == (out_infofile, 3)
    out, err = capsys.readouterr()
    # Check logs
    assert ("Generating distribution of L90 and #contigs graphs.") in out

    # Check output files present
    protdir = os.path.join(GENEPATH, "Proteins")
    assert len(os.listdir(protdir)) == 3
    gffdir = os.path.join(GENEPATH, "gff3")
    assert len(os.listdir(gffdir)) == 3
    lstdir = os.path.join(GENEPATH, "LSTINFO")
    assert len(os.listdir(lstdir)) == 3

    # Check genomes are renamed as expected, and with expected L90/nbcont values
    exp_lstinfo = os.path.join(EXP_DIR, "exp_LSTINFO-test-main-frominfo.lst")
    res_lstinfo = os.path.join(GENEPATH, "LSTINFO-lstinfo.lst")
    assert tutil.compare_order_content(exp_lstinfo, res_lstinfo)


def test_main_novalid_genome_frominfo(capsys):
    """
    Test that when, in the list file, all genomes are wrong (do not correspond to
    filenames in the given dbpath), it closes the program with an error message.
    """
    listfile = None
    dbpath = None
    name = "TOTO"
    date = "1205"
    infofile = os.path.join(TEST_DIR, "lstinfo-no-genome.lst")
    with pytest.raises(SystemExit):
        annot.main("cmd", listfile, dbpath, GENEPATH, name, date, from_info=infofile,
                   prodigal_only=True)
    out, err = capsys.readouterr()
    # Check logs
    assert ("No genome listed in test/data/annotate/test_files/lstinfo-no-genome.lst "
            "was found.") in err

    # Check output folders not created
    protdir = os.path.join(GENEPATH, "Proteins")
    assert not os.path.isdir(protdir)
    gffdir = os.path.join(GENEPATH, "gff3")
    assert not os.path.isdir(gffdir)
    lstdir = os.path.join(GENEPATH, "LSTINFO")
    assert not os.path.isdir(lstdir)

    # Check tmp_files is empty
    tmp = os.path.join(GENEPATH, "tmp_files")
    assert len(os.listdir(tmp)) == 0



#!/usr/bin/env python3
# coding: utf-8

# ###############################################################################
# This file is part of PanACOTA.                                                #
#                                                                               #
# Authors: Amandine Perrin                                                      #
# Copyright © 2018-2020 Institut Pasteur (Paris).                               #
# See the COPYRIGHT file for details.                                           #
#                                                                               #
# PanACOTA is a software providing tools for large scale bacterial comparative  #
# genomics. From a set of complete and/or draft genomes, you can:               #
#    -  Do a quality control of your strains, to eliminate poor quality         #
# genomes, which would not give any information for the comparative study       #
#    -  Uniformly annotate all genomes                                          #
#    -  Do a Pan-genome                                                         #
#    -  Do a Core or Persistent genome                                          #
#    -  Align all Core/Persistent families                                      #
#    -  Infer a phylogenetic tree from the Core/Persistent families             #
#                                                                               #
# PanACOTA is free software: you can redistribute it and/or modify it under the #
# terms of the Affero GNU General Public License as published by the Free       #
# Software Foundation, either version 3 of the License, or (at your option)     #
# any later version.                                                            #
#                                                                               #
# PanACOTA is distributed in the hope that it will be useful, but WITHOUT ANY   #
# WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS     #
# FOR A PARTICULAR PURPOSE. See the Affero GNU General Public License           #
# for more details.                                                             #
#                                                                               #
# You should have received a copy of the Affero GNU General Public License      #
# along with PanACOTA (COPYING file).                                           #
# If not, see <https://www.gnu.org/licenses/>.                                  #
# ###############################################################################

"""
Functions to infer a phylogenetic tree with quicktree

@author gem
June 2017
"""

from Bio import AlignIO
import os
import logging

from PanACoTA import utils

logger = logging.getLogger("tree.quicktree")


def run_tree(alignfile, boot, outdir, *args, **kwargs):
    """
    Run quicktree for the given alignment file and options

    Parameters
    ----------
    alignfile: str
        Path to file containing alignments of persistent families grouped by genome
    boot: int or None
        Number of bootstraps to compute. None if no bootstrap asked
    outdir: str or None
        Path to the tree file that must be created
    args: tuple
        Used to be compatible with the 'run_tree' function of other softs like fastME and
        fastTree which require more arguments like the DNA substitution model, the number of
        threads to use, etc.
    kwargs: dict
        Used to be compatible with the 'run_tree' function of other softs like fastME and
        fastTree which require more arguments like the DNA substitution model, the number of
        threads to use, etc.
    """
    align_name = os.path.basename(alignfile)
    align_stock = os.path.join(outdir, align_name + ".stockholm")
    convert2stockholm(alignfile, align_stock)
    run_quicktree(align_stock, boot, outdir)


def convert2stockholm(infile, outfile):
    """
    Input alignment is in fasta format. Input of quicktree must be in stockholm format.
    Convert it here.

    Parameters
    ----------
    infile: str
        Path to file containing alignments in fasta
    outfile: str
        Path to file which will contain the alignments converted to Stockholm format
    """
    if os.path.isfile(outfile):
        logger.info("Stockholm alignment file already existing.")
        logger.warning(("The Stockholm alignment file {} already exists. The program "
                        "will use it instead of re-converting {}.").format(outfile, infile))
        return
    logger.info("Converting fasta alignment to stockholm format.")
    with open(infile, 'r') as input_handle, open(outfile, 'w') as output_handle:
        alignments = AlignIO.parse(input_handle, "fasta")
        AlignIO.write(alignments, output_handle, "stockholm")


def run_quicktree(alignfile, boot, outdir):
    """
    Run quicktree on the given alignment.

    Parameters
    ----------
    alignfile: str
        Path to file containing alignments of persistent families grouped by genome,
        in Stockholm format
    boot: int or None
        Number of bootstraps to compute. None if no bootstrap asked
    outdir: str or None
        Path to the tree file that must be created
    """
    logger.info("Running Quicktree...")
    bootinfo = ""

    # Get bootstrap information
    if boot:
        bootinfo = f"-boot {boot}"
    # Get output filename and logfile name
    align_name = os.path.basename(alignfile)
    logfile = os.path.join(outdir, align_name + ".quicktree.log")
    treefile = os.path.join(outdir, align_name + ".quicktree_tree.nwk")
    cmd = f"quicktree -in a -out t {bootinfo} {alignfile}"
    outfile = open(treefile, "w")
    logfilef = open(logfile, "w")
    error = (f"Problem while running quicktree. See log file ({logfile}) for "
             "more information.")
    logger.details(cmd)
    utils.run_cmd(cmd, error, stdout=outfile, eof=True, logger=logger, stderr=logfilef)



#!/usr/bin/env python3
# coding: utf-8

"""
This script contains functions used by several tests.
"""

import hashlib


def get_file_hash(filepath):
    """
    Returns the MD5 hash of a file 'filepath'. This is the signature of a file, used to check data
    integrity of this file.
    """
    fileobj = open(filepath, 'rb')
    m = hashlib.md5()
    while True:
        d = fileobj.read(2**20)
        if not d:
            break
        m.update(d)
    fileobj.close()
    return m.hexdigest()


def compare_files_bin(file1, file2):
    """
    Checks if the two files are the same or not, by comparing their signatures.
    """
    sig_file1 = get_file_hash(file1)
    sig_file2 = get_file_hash(file2)
    return sig_file1 == sig_file2


def compare_file_content(file1, file2):
    """
    Check that the 2 files have the same content, but no necessarily in the same order
    """
    with open(file1, "r") as f1:
        lines_f1 = f1.readlines()
        lines_f1 = [l.strip() for l in lines_f1]
        # Remove empty lines at the end of the file
        while lines_f1[-1].strip() == "":
            lines_f1 = lines_f1[:-1]

    with open(file2, "r") as f2:
        lines_f2 = f2.readlines()
        lines_f2 = [l.strip() for l in lines_f2]
        while lines_f2[-1].strip() == "":
            lines_f2 = lines_f2[:-1]
        if len(lines_f1) != len(lines_f2):
            print(f"not same number of lines in {file1} and {file2}")
            return False
        for line2 in lines_f2:
            if line2 not in lines_f1:
                print(f"'{line2}' not found")
                return False
    return True


def compare_order_content(file1, file2):
    """
    Check that the 2 files have exactly the same content, in the same order
    """
    with open(file1, "r") as f1:
        lines_f1 = f1.readlines()
        # Remove empty lines at the end of the file
        while lines_f1[-1].strip() == "":
            lines_f1 = lines_f1[:-1]
    with open(file2, "r") as f2:
        lines_f2 = f2.readlines()
        # Remove empty lines at the end of the file
        while lines_f2[-1].strip() == "":
            lines_f2 = lines_f2[:-1]

    # Compare number of lines
    if len(lines_f1) != len(lines_f2):
        print(f"not same number of lines in {file1} and {file2}")
        return False

    # Compare content of lines, in order
    for line1, line2 in zip(lines_f1, lines_f2):
        if line1.split() != line2.split():
            print(f"'{line1}' vs '{line2}' do not correspond.")
            return False
    return True


def compare_file_to_list(file1, exp):
    """
    Check that the file contains the same lines as the list 'exp', in any order
    """
    missing = []
    with open(file1, "r") as outf:
        for line in outf:
            missing.append(line.strip())
    if len(missing) != len(exp):
        print(f"{len(missing)} lines in {file1}, {len(exp)} lines expected")
        return False
    if set(missing) != set(exp):
        print("Not same lines")
        return False
    return True



#!/usr/bin/env python3

"""
Functional tests for PanACoTA 'all' module
"""
import pytest
import os
import shutil
import glob
import argparse

import PanACoTA.subcommands.all_modules as allm

DATADIR = os.path.join("test", "data", "all")
GENEPATH = os.path.join(DATADIR, "generated_by_func-tests")

@pytest.fixture(autouse=True)
def setup_teardown_module():
    """
    Remove log files at the end of this test module

    Before each test:
    - init logger
    - create directory to put generated files

    After:
    - remove all log files
    - remove directory with generated results
    """
    os.mkdir(GENEPATH)
    print("setup")

    yield
    shutil.rmtree(GENEPATH)
    print("teardown")


def test_main_default_qc_only():
    """
    Test downloading 104099 genomes and analysis until QC, with verbose==2
    """
    cmd = "cmd"
    outdir = os.path.join(GENEPATH, "main_all_default_qc")

    # Common arguments: outdir, threads, verbose, quiet
    args_all = (outdir, 1, 2, False)
    # args for prepare:
    # NCBI_species_taxid (int), NCBI_species (str), levels (str), tmp_dir (str),
    # norefseq (bool), db_dir (str), only_mash (bool), info_file (str), l90 (int),
    # nbcont (int), cutn (int), min_dist (float), max_dist (float)
    args_prepare = ("104099", "", "", "all", "refseq", "", False, "", False, "", 100, 999, 5, 1e-4, 0.06)
    # args for annotate:
    # name (str), qc_only (bool), date (str), prodigal_only (bool), small (bool)
    args_annot = ("TEST", True, "2101", False, False)
    # args for pangenome:
    # min_id (float), clust_mode (int), spe_dir (str), outfile (str)
    args_pan = (0.8, 1, "", "")
    # args for corepers
    # tol (float), mixed (bool), multi (bool), floor (bool)
    args_corepers = (1, False, False, False)
    #  args for tree module
    #  soft (str), model (str), boot (bool), write_boot (bool), memory (str), fast (bool)
    args_tree = ("iqtree2", "GTR", False, False, "", True)

    # Run 'all' module
    out = allm.main(cmd, args_all, args_prepare, args_annot, args_pan, args_corepers, args_tree)
    assert out == "QC_only done"
    # Check that there are 3 log files (log, err and details)
    log_files = glob.glob(os.path.join(outdir, "*log*"))
    assert len(log_files) == 3
    # Check that there are 5 files/folders : 3 logs + 2 result folders
    assert len(os.listdir(outdir)) == 5
    # Check result folder names
    prep_dir = os.path.join(outdir, "1-prepare_module")
    annot_dir = os.path.join(outdir, "2-annotate_module")
    assert os.path.isdir(prep_dir)
    assert os.path.isdir(annot_dir)
    # Check presence of some key files for prepare module
    assert(len(glob.glob(os.path.join(prep_dir, "*log*")))) == 3
    lst1 = os.path.join(prep_dir, "LSTINFO-104099-filtered-0.0001_0.06.txt")
    ass1 = os.path.join(prep_dir, "assembly_summary-104099.txt") 
    assert os.path.isfile(lst1)
    assert os.path.isfile(ass1)
    assert os.path.isdir(os.path.join(prep_dir, "refseq"))
    # Check presence of key files in prepare module
    assert(len(glob.glob(os.path.join(annot_dir, "*log*")))) == 3
    lst2 = os.path.join(annot_dir, "ALL-GENOMES-info-LSTINFO-104099-filtered-0.0001_0.06.lst")
    png1 = os.path.join(annot_dir, "QC_L90-LSTINFO-104099-filtered-0.0001_0.06.png")
    png2 = os.path.join(annot_dir, "QC_nb-contigs-LSTINFO-104099-filtered-0.0001_0.06.png")
    disc = os.path.join(annot_dir, "discarded-LSTINFO-104099-filtered-0.0001_0.06.lst")
    for f in (lst2, png1, png2, disc):
        assert os.path.isfile(f)


def test_main_norefseq():
    """
    Test with norefseq (4 genomes given, giving 13 families).
    """
    cmd = "cmd"
    outdir = os.path.join(GENEPATH, "main_all_default_qc")

    # Common arguments: outdir, threads, verbose, quiet
    args_all = (outdir, 1, 0, False)
    # args for prepare:
    # NCBI_species_taxid (int), NCBI_species (str), levels (str), tmp_dir (str),
    # norefseq (bool), db_dir (str), only_mash (bool), info_file (str), l90 (int),
    # nbcont (int), cutn (int), min_dist (float), max_dist (float)
    # db_dir = "test/data/pangenome/test_files/example_db/Replicons"
    # db_dir = "104099/Database_init"
    db_dir = os.path.join(DATADIR, "genomes")
    args_prepare = ("104099", "", "", "all", "refseq", "", True, db_dir, False, "", 100, 999, 5, 1e-4, 1)
    # args for annotate:
    # name (str), qc_only (bool), date (str), prodigal_only (bool), small (bool)
    args_annot = ("TEST", False, "2101", True, False)
    # args for pangenome:
    # min_id (float), clust_mode (int), spe_dir (str), outfile (str)
    args_pan = (0.8, 1, "", "")
    # args for args_corepers
    # tol (float), mixed (bool), multi (bool), floor (bool)
    args_corepers = (1, False, False, False)
    #  args for tree module
    #  soft (str), model (str), boot (bool), write_boot (bool), memory (str), fast (bool)
    args_tree = ("iqtree2", "GTR", False, False, "", True)

    # Run 'all' module
    out = allm.main(cmd, args_all, args_prepare, args_annot, args_pan, args_corepers, args_tree)
    assert out == 0
    # Check that there are 2 log files (log, err)
    log_files = glob.glob(os.path.join(outdir, "*log*"))
    assert len(log_files) == 2
    # Check that there are 5 files/folders : 2 logs + 4 result folders
    assert len(os.listdir(outdir)) == 8
    # Check result folder names
    prep_dir = os.path.join(outdir, "1-prepare_module")
    annot_dir = os.path.join(outdir, "2-annotate_module")
    pan_dir = os.path.join(outdir, "3-pangenome_module")
    core_dir = os.path.join(outdir, "4-corepers_module")
    ali_dir = os.path.join(outdir, "5-align_module")
    tree_dir = os.path.join(outdir, "6-tree_module")
    for d in (prep_dir, annot_dir, pan_dir, core_dir, ali_dir, tree_dir):
        assert os.path.isdir(d)
    # CHECK PREPARE
    # Check presence of some key files for prepare module
    assert(len(glob.glob(os.path.join(prep_dir, "*log*")))) == 3
    lst1 = os.path.join(prep_dir, "LSTINFO-104099-filtered-0.0001_1.txt")
    assert os.path.isfile(lst1)
    # CHECK ANNOTATE
    # Check presence of key files in prepare module
    assert(len(glob.glob(os.path.join(annot_dir, "*log*")))) == 3
    lst2 = os.path.join(annot_dir, "LSTINFO-LSTINFO-104099-filtered-0.0001_1.lst")
    png1 = os.path.join(annot_dir, "QC_L90-LSTINFO-104099-filtered-0.0001_1.png")
    png2 = os.path.join(annot_dir, "QC_nb-contigs-LSTINFO-104099-filtered-0.0001_1.png")
    for f in (lst2, png1, png2):
        assert os.path.isfile(f)
    # Check presence of proteins, genes, gff3, replicons and lstinfo folders, and 4 genomes in each
    prot_fold = os.path.join(annot_dir, "Proteins")
    gen_fold = os.path.join(annot_dir, "Genes")
    rep_fold = os.path.join(annot_dir, "Replicons")
    gff_fold = os.path.join(annot_dir, "gff3")
    lstinfo_fold = os.path.join(annot_dir, "LSTINFO")
    for d in (gen_fold, rep_fold, gff_fold, lstinfo_fold):
        assert os.path.isdir(d)
        files = os.listdir(d)
        assert len(files) == 4
    assert os.path.isdir(prot_fold)
    assert len(os.listdir(prot_fold)) == 5  # 4 genomes + concatenated DB
    # CHECK PAN
    assert(len(glob.glob(os.path.join(pan_dir, "PanACoTA*log*")))) == 3
    assert(len(glob.glob(os.path.join(pan_dir, "PanGenome-TEST*")))) == 5
    # CHECK COREPERS
    assert(len(glob.glob(os.path.join(core_dir, "PanACoTA*log*")))) == 2
    core_file = glob.glob(os.path.join(core_dir, "PersGenome*"))
    assert len(core_file) == 1
    # CHECK PAN
    assert(len(glob.glob(os.path.join(ali_dir, "PanACoTA*log*")))) == 3
    assert os.path.isdir(os.path.join(ali_dir, "Align-TEST_4"))
    assert os.path.isdir(os.path.join(ali_dir, "List-TEST_4"))
    assert os.path.isdir(os.path.join(ali_dir, "Phylo-TEST_4"))
    assert os.path.isfile(os.path.join(ali_dir, "Phylo-TEST_4", "TEST_4.grp.aln"))
    # CHECK TREE
    assert(len(glob.glob(os.path.join(tree_dir, "PanACoTA*log*")))) == 3
    assert os.path.isfile(os.path.join(tree_dir, "TEST_4.grp.aln.iqtree_tree.iqtree"))
    assert os.path.isfile(os.path.join(tree_dir, "TEST_4.grp.aln.iqtree_tree.treefile"))


def test_main_from_parse():
    """
    same test (from norefseq), but from 'main_from_parse' function, and with verbose == debug
    """
    args = argparse.Namespace()
    args.argv = ["all", "test_all_modules"]
    # common params
    outdir = os.path.join(GENEPATH, "from_parse")
    args.outdir = outdir
    args.threads = 1
    args.verbose = 15
    args.quiet = False
    # prepare params
    args.ncbi_species_name = ""
    args.ncbi_species_taxid = "104099"
    args.ncbi_taxid = ""
    args.levels = ""
    args.ncbi_section = "refseq"
    args.tmp_dir = ""
    args.norefseq = True
    args.db_dir = os.path.join(DATADIR, "genomes")
    args.only_mash = False
    args.info_file = ""
    args.l90 = 100
    args.nbcont = 999
    args.cutn = 5
    args.min_dist = 1e-4
    args.max_dist = 1
    # annotate params
    args.qc_only = False
    args.date = "2101"
    args.prodigal_only = False
    args.small = False
    args.name = "TEST"
    # pangenome params
    args.min_id = 0.8
    args.clust_mode = 1
    args.spedir = ""
    args.outfile = ""
    # corepers params
    args.tol = 1
    args.mixed = False
    args.multi = False
    args.floor = False
    # params tree
    args.soft = "iqtree2"
    args.model = "GTR"
    args.boot = False
    args.write_boot = False 
    args.memory = ""
    args.fast = True

    allm.main_from_parse(args)

    # Check that there are 2 log files (log, err)
    log_files = glob.glob(os.path.join(outdir, "*log*"))
    assert len(log_files) == 4
    # Check that there are 5 files/folders : 4 logs + 6 result folders
    assert len(os.listdir(outdir)) == 10
    # Check result folder names
    prep_dir = os.path.join(outdir, "1-prepare_module")
    annot_dir = os.path.join(outdir, "2-annotate_module")
    pan_dir = os.path.join(outdir, "3-pangenome_module")
    core_dir = os.path.join(outdir, "4-corepers_module")
    ali_dir = os.path.join(outdir, "5-align_module")
    tree_dir = os.path.join(outdir, "6-tree_module")
    for d in (prep_dir, annot_dir, pan_dir, core_dir, ali_dir, tree_dir):
        assert os.path.isdir(d)
    # CHECK PREPARE
    # Check presence of some key files for prepare module
    assert(len(glob.glob(os.path.join(prep_dir, "*log*")))) == 4
    lst1 = os.path.join(prep_dir, "LSTINFO-104099-filtered-0.0001_1.txt")
    assert os.path.isfile(lst1)
    # CHECK ANNOTATE
    # Check presence of key files in prepare module
    assert(len(glob.glob(os.path.join(annot_dir, "*log*")))) == 4
    lst2 = os.path.join(annot_dir, "LSTINFO-LSTINFO-104099-filtered-0.0001_1.lst")
    png1 = os.path.join(annot_dir, "QC_L90-LSTINFO-104099-filtered-0.0001_1.png")
    png2 = os.path.join(annot_dir, "QC_nb-contigs-LSTINFO-104099-filtered-0.0001_1.png")
    for f in (lst2, png1, png2):
        assert os.path.isfile(f)
    # Check presence of proteins, genes, gff3, replicons and lstinfo folders, and 4 genomes in each
    prot_fold = os.path.join(annot_dir, "Proteins")
    gen_fold = os.path.join(annot_dir, "Genes")
    rep_fold = os.path.join(annot_dir, "Replicons")
    gff_fold = os.path.join(annot_dir, "gff3")
    lstinfo_fold = os.path.join(annot_dir, "LSTINFO")
    for d in (gen_fold, rep_fold, gff_fold, lstinfo_fold):
        assert os.path.isdir(d)
        files = os.listdir(d)
        assert len(files) == 4
    assert os.path.isdir(prot_fold)
    assert len(os.listdir(prot_fold)) == 5  # 4 genomes + concatenated DB
    # CHECK PAN
    assert(len(glob.glob(os.path.join(pan_dir, "PanACoTA*log*")))) == 4
    assert(len(glob.glob(os.path.join(pan_dir, "PanGenome-TEST*")))) == 5
    # CHECK COREPERS
    assert(len(glob.glob(os.path.join(core_dir, "PanACoTA*log*")))) == 4
    core_file = glob.glob(os.path.join(core_dir, "PersGenome*"))
    assert len(core_file) == 1
    # CHECK PAN
    assert(len(glob.glob(os.path.join(ali_dir, "PanACoTA*log*")))) == 4
    assert os.path.isdir(os.path.join(ali_dir, "Align-TEST_4"))
    assert os.path.isdir(os.path.join(ali_dir, "List-TEST_4"))
    assert os.path.isdir(os.path.join(ali_dir, "Phylo-TEST_4"))
    assert os.path.isfile(os.path.join(ali_dir, "Phylo-TEST_4", "TEST_4.grp.aln"))
    # CHECK TREE
    assert(len(glob.glob(os.path.join(tree_dir, "PanACoTA*log*")))) == 4
    assert os.path.isfile(os.path.join(tree_dir, "TEST_4.grp.aln.iqtree_tree.iqtree"))
    assert os.path.isfile(os.path.join(tree_dir, "TEST_4.grp.aln.iqtree_tree.treefile"))


#!/usr/bin/env python3
# coding: utf-8

# ###############################################################################
# This file is part of PanACOTA.                                                #
#                                                                               #
# Authors: Amandine Perrin                                                      #
# Copyright © 2018-2020 Institut Pasteur (Paris).                               #
# See the COPYRIGHT file for details.                                           #
#                                                                               #
# PanACOTA is a software providing tools for large scale bacterial comparative  #
# genomics. From a set of complete and/or draft genomes, you can:               #
#    -  Do a quality control of your strains, to eliminate poor quality         #
# genomes, which would not give any information for the comparative study       #
#    -  Uniformly annotate all genomes                                          #
#    -  Do a Pan-genome                                                         #
#    -  Do a Core or Persistent genome                                          #
#    -  Align all Core/Persistent families                                      #
#    -  Infer a phylogenetic tree from the Core/Persistent families             #
#                                                                               #
# PanACOTA is free software: you can redistribute it and/or modify it under the #
# terms of the Affero GNU General Public License as published by the Free       #
# Software Foundation, either version 3 of the License, or (at your option)     #
# any later version.                                                            #
#                                                                               #
# PanACOTA is distributed in the hope that it will be useful, but WITHOUT ANY   #
# WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS     #
# FOR A PARTICULAR PURPOSE. See the Affero GNU General Public License           #
# for more details.                                                             #
#                                                                               #
# You should have received a copy of the Affero GNU General Public License      #
# along with PanACOTA (COPYING file).                                           #
# If not, see <https://www.gnu.org/licenses/>.                                  #
# ###############################################################################

"""
Util functions and classes.


@author gem
April 2017
"""

import os
import sys
import re
import glob
import subprocess
import shutil
import shlex
import progressbar

# Logging
import logging
from logging.handlers import RotatingFileHandler
from colorlog import ColoredFormatter

try:
    import cPickle as pickle
except:
    try:
        import _pickle as pickle
    except:  # pragma: no cover
        import pickle


def init_logger(logfile_base, level, name, log_details=False, verbose=0, quiet=False):
    """
    Create logger and its handlers, and set them to the given level

    level hierarchy: ``CRITICAL > ERROR > WARNING > INFO > DETAILS > DEBUG``

    Messages from all levels are written in 'logfile'.log

    Messages for levels less than WARNING (only INFO and DEBUG) written to stdout

    Messages for levels equal or higher than WARNING written to stderr

    Messages for levels equal or higher than WARNING written in `logfile`.log.err


    Parameters
    ----------
    logfile_base : str
        base of filename to use for logs. Will add '.log', '.log.details' and '.log.err' for\
        the 3 log files created
    level : int
        minimum level that must be considered.
    name : str or None
        if we need to name the logger (used for tests)
    log_details : bool
        if True, force creation of .log.details file. Otherwise, just create
        it if needed according to level

    verbose : int
        be more verbose:
        default (0): info in stdout, error and more in stderr ;
        info and more in *.log ; warning and more in *.log.err
        1 = add warnings in stderr
        2 = like 1 + add details to stdout (by default only INFO) + add details to *.log.details
        >15: add debug to stdout and create *.log.debug with all levels
    quiet : bool
        True if nothing must be sent to stdout/stderr, False otherwise
    """
    import time

    # time when soft is launched
    time_start = time.strftime("_%y-%m-%d_%H-%m-%S")

    # create logger
    logger = logging.getLogger(name)

    # Determine logfile names
    logfile = logfile_base + ".log"
    if os.path.isfile(logfile):
        logfile = logfile_base + "-" + time_start + ".log"
    errfile = logfile_base + ".log.err"
    if os.path.isfile(errfile):
        errfile = logfile_base + "-" + time_start + ".log.err"
    detailfile = logfile_base + ".log.details"
    if os.path.isfile(detailfile):
        detailfile = logfile_base + "-" + time_start + ".log.details"
    debugfile = logfile_base + ".log.debug"
    if os.path.isfile(debugfile):
        debugfile = logfile_base + "-" + time_start + ".log.debug"

    # Create a new logging level: details (between info and debug)
    # Used to add details to the log file, but not to stdout, while still having
    # the possibility to put debug messages, used only for development.
    def details(self, message, *args, **kws):
        """
        Define a new log level: details
        """
        if self.isEnabledFor(logging.DETAIL):
            self._log(logging.DETAIL, message, args, **kws)
    logging.addLevelName(detail_lvl(), "DETAIL")
    logging.DETAIL = detail_lvl()
    logging.Logger.details = details

    # set level of logger
    logger.setLevel(logging.DEBUG)

    # create formatter for log messages:
    # "timestamp :: level :: message"
    # (add :: %(name)s  to add the logger name)
    my_format = '[%(asctime)s] :: %(levelname)s :: %(message)s'
    formatter_file = logging.Formatter(my_format, '%Y-%m-%d %H:%M:%S')
    my_format_stream = '%(log_color)s  * [%(asctime)s] : %(levelname)s %(reset)s %(message)s'
    formatter_stream = ColoredFormatter(my_format_stream, datefmt='%Y-%m-%d %H:%M:%S',
                                        log_colors={'DEBUG' : 'cyan',
                                                    'INFO' : 'green',
                                                    'DETAIL' : 'cyan',
                                                    'WARNING' : 'yellow',
                                                    'ERROR' : 'red',
                                                    'CRITICAL' : 'red',
                                                    })



    # Create handler 1: writing to 'logfile'. mode 'write', max size = 1Mo.
    # If logfile is 1Mo, it is renamed to logfile.1, and next logs are still
    # written to logfile. Then, logfile.1 is renamed to logfile.2, logfile to
    # logfile.1 etc. We allow maximum 5 log files.
    # logfile contains everything from INFO level (INFO, WARNING, ERROR)
    logfile_handler = RotatingFileHandler(logfile, 'w', 10000000, 5)
    # set level to the same as the logger level
    logfile_handler.setLevel(logging.INFO)
    logfile_handler.setFormatter(formatter_file)  # add formatter
    logger.addHandler(logfile_handler)  # add handler to logger

    # Create handler 2: errfile. Write only warnings and errors
    errfile_handler = RotatingFileHandler(errfile, 'w', 10000000, 5)
    errfile_handler.setLevel(logging.WARNING)
    errfile_handler.setFormatter(formatter_file)  # add formatter
    logger.addHandler(errfile_handler)  # add handler to logger


    # Create handler 3: detailsfile. Write everything to this file, except debug
    # Create it only if:
    # - level is <= info (for modules which have no details, so detailsfile is the same as
    # logfile)
    # - details==True force creation of detailsfile
    # - quiet==True nothing in stdout, put all log files so that user can check
    if level < logging.INFO or quiet or log_details:
        detfile_handler = RotatingFileHandler(detailfile, 'w', 10000000, 5)
        detfile_handler.setLevel(logging.DETAIL)
        detfile_handler.setFormatter(formatter_file)  # add formatter
        logger.addHandler(detfile_handler)  # add handler to logger

    # Formats for detailed log files
    my_format_debug = '[%(asctime)s] :: %(levelname)s (from %(name)s logger) :: %(message)s'
    formatter_file_debug = logging.Formatter(my_format_debug, '%Y-%m-%d %H:%M:%S')

    # Create handler 4: debug file. Write everything
    if level < logging.DETAIL:
        debugfile_handler = RotatingFileHandler(debugfile, 'w', 10000000, 5)
        debugfile_handler.setLevel(logging.DEBUG)
        debugfile_handler.setFormatter(formatter_file_debug)  # add formatter
        logger.addHandler(debugfile_handler)  # add handler to logger

    # If not quiet, add handlers for stdout and stderr
    if not quiet:
        # Create handler 4: write to stdout
        stream_handler = logging.StreamHandler(sys.stdout)
        # By default, write everything
        stream_handler.setLevel(logging.DEBUG)
        # BUT: don't write messages >= WARNING (warning, error, critical)
        stream_handler.addFilter(LessThanFilter(logging.WARNING))
        # if not verbose (level 0 or 1): only put info in stdout (remove details and debug)
        if verbose < 2:
            stream_handler.addFilter(NoLevelFilter(logging.DETAIL))
            stream_handler.addFilter(NoLevelFilter(logging.DEBUG))
        stream_handler.setFormatter(formatter_stream)
        logger.addHandler(stream_handler)  # add handler to logger

        # Create handler 5: write to stderr
        err_handler = logging.StreamHandler(sys.stderr)

        if verbose > 0:
            err_handler.setLevel(logging.WARNING)  # write all messages >= WARNING
        else:
            err_handler.setLevel(logging.ERROR)  # write all messages >= ERROR
        err_handler.setFormatter(formatter_stream)
        logger.addHandler(err_handler)  # add handler to logger
    logger.propagate = False
    return logfile, logger


class LessThanFilter(logging.Filter):
    """
    When using log, when a level is set to a handler, it is a minimum level. All
    levels higher than it will be printed. If you want to print only until
    a given level (no levels higher than the specified one), use this class like this:
    handler.addFilter(LessThanFilter(level))
    """

    def __init__(self, level):
        self._level = level
        logging.Filter.__init__(self)

    def filter(self, rec):
        """
        Function to decide if given log has to be logged or not, according to its level

        Parameters
        ----------
        rec : current record handled by logger

        Returns
        -------
        bool
            True if level of current log is less than the defined limit, False otherwise
        """
        return rec.levelno < self._level


class NoLevelFilter(logging.Filter):
    """
    When using log, specify a given level that must not be taken into account by the handler.
    This is used for the stdout handler. We want to print, by default,
    DEBUG (for development use) and INFO levels, but not DETAILS level (which is between
    DEBUG and INFO). We want to print DETAIL only if verbose option was set
    """

    def __init__(self, level):
        self._level = level
        logging.Filter.__init__(self)

    def filter(self, rec):
        """
        Function to decide if given log has to be logged or not, according to its level

        Parameters
        ----------
        rec : current record handled by logger

        Returns
        -------
        bool
            True if level of current log is different from forbidden level, False if it is the same
        """
        return rec.levelno != self._level


def check_installed(cmd):
    """
    Check if the command 'cmd' is in $PATH and can then be executed

    Parameters
    ----------
    cmd : str
        command to run

    Returns
    -------
    bool
        True if installed, False otherwise
    """
    torun = "which " + cmd
    trying = subprocess.Popen(shlex.split(torun), stdout=subprocess.PIPE)
    out, _ = trying.communicate()
    if trying.returncode == 0:
        if os.path.isfile(out.strip()):
            return True
    return False


def run_cmd(cmd, error, eof=False, **kwargs):
    """
    Run the given command line. If the return code is not 0, print error message.
    if eof (exit on fail) is True, exit program if error code is not 0.

    Parameters
    ----------
    cmd : str
        command to run
    error : str
        error message to print if error while running command
    eof : bool
        True: exit program if command failed, False: do not exit even if command fails
    kwargs : Object
        Can provide a logger, stdout and/or stderr streams

    Returns
    -------
    subprocess.Popen
        returns object of subprocess call (has attributes returncode, pid, communicate etc.)

    """
    if "logger" not in kwargs:
        logger = logging.getLogger("utils.run_cmd")
    else:
        logger = kwargs["logger"]
    if "stdout" not in kwargs:
        kwargs["stdout"] = None
    if "stderr" not in kwargs:
        kwargs["stderr"] = None
    try:
        call = subprocess.Popen(shlex.split(cmd), stdout=kwargs["stdout"],
                                stderr=kwargs["stderr"])
        call.wait()
        retcode = call.returncode
    except OSError:
        logger.error(f"error: command '>{cmd}' is not possible.")
        if eof:
            sys.exit(1)
        else:
            return 1
    if retcode != 0:
        logger.error(error)
        if eof:
            sys.exit(retcode)
    return call


def plot_distr(values, limit, title, text, logger):
    """
    Plot histogram of given 'values', and add a vertical line corresponding to the chosen
    'limit' and return the mpl figure

    Parameters
    ----------
    values : list
        list of values
    limit : int
        limit for which a vertical line must be drawn
    title : str
        Title to give to plot
    text : str
        text to write near the vertical line representing the limit
    logger : logging.Logger
        logger object to write log information

    Returns
    -------
    matplotlib.figure.Figure
        figure generated
    """
    import math
    import numpy as np
    import matplotlib
    matplotlib.use('AGG')
    from matplotlib import pyplot as plt
    plt.close("all")
    fig = plt.figure(figsize=(10, 7))
    ax = fig.add_subplot(1, 1, 1)
    max_x = max(values)
    # if too many values, group them to have less bins in the histogram.
    # Put 'group_values' values in each bin ->
    # if less than 300 values, 1 value per bin, otherwise more values per bin
    group_values = int(max_x / 300) + 1
    dec_ax = math.exp(0.001 * max_x) - 1
    dec_text = 3 * dec_ax
    bins = np.arange(0, max_x + 2 * group_values, group_values) - 0.5
    ax.hist(values, bins=bins, edgecolor="black", color="blue")
    ax.set_xlim(0.5, max_x + 0.5 * group_values)
    ax.axvline(x=limit + 0.5 * group_values + dec_ax, color="r")
    ax.text(x=limit + 0.5 * group_values + dec_text, y=plt.ylim()[1] / 2,
            s=text + " " + str(limit), color="r", rotation=90)
    ax.set_title(title)
    return fig


def write_warning_skipped(skipped, do_format=False, prodigal_only=False, logfile=""):
    """
    At the end of the script, write a warning to the user with the names of the genomes
    which had problems with prokka.

    Parameters
    ----------
    skipped : list
        list of genomes with problems
    do_format : bool
        if False, genomes were not skipped because of format step, but before that.\
        if True, they were skipped because of format
    prodigal_only : bool
        if False: used prokka to annotate
        if True: used prodigal to annotate
    """
    if not prodigal_only:
        soft = "prokka"
    else:
        soft = "prodigal"
    logger = logging.getLogger("utils")
    list_to_write = "\n".join(["\t- " + genome for genome in skipped])
    if not do_format:
        logger.info(f"WARNING: Some genomes could not be annotated. See {soft}")
        logger.warning(f"{soft} had problems while annotating some genomes, or "
                    "did not find any gene. Hence, they are not formatted, and absent "
                    "from your output database. Please look at the "
                    "current error log "
                    "(<output_directory>/PanACoTA-annotate_list_genomes[-date].log.err) to get more "
                    "information on the problems. Here are those "
                    f"genomes:\n{list_to_write}")
    else:
        logger.info(f"WARNING: Some genomes could not be formatted. See {logfile}")
        logger.warning((f"Some genomes were annotated by {soft}, but could not be formatted, "
                        "and are hence absent from your output database. Please look at "
                        "'<output_directory>/PanACoTA-annotate_list_genomes[-date].log.err' and "
                        ".details files to get more information about why they could not be "
                        f"formatted.\n{list_to_write}"))


def write_genomes_info(genomes, kept_genomes, list_file, res_path, qc=False):
    """
    Write the list of genomes discarded to a file (qc=False), so that users can
    keep a trace of them, with their information (nb contigs, L90 etc.)

    If qc=True, we stop after QC.
    -> Write the list of genomes that would be kept for annotation with all
    their information (L90, size, #contig)


    Parameters
    ----------
    genomes : dict
        {genome: [gembase_start_name, orig_seq_file, to_annotate_seq_file,
                  genome_size, nb_contigs, L90]}
    kept_genomes : list
        list of genomes kept
    list_file : str
        path to input file containing the list of genomes
    res_path : str
        folder where results must be saved
    qc : bool
        * True: called only if QC only. Name this file info-genomes-<list_file>.txt to put
        information on genomes that would be annotated if not QC only
        * otherwise (False), called in any case. Name this file discarded-<list_file>.txt
        and write all discarded genomes, whether sequences kept are next annotated or not
        => columns: orig_name, to_annotate, gsize, nb_conts, L90
    """
    logger = logging.getLogger("utils")
    # number of genomes discarded
    nb_disc = len(genomes) - len(kept_genomes)
    # Log number of genomes discarded.
    if not qc and nb_disc < 2:
        logger.info(f"{nb_disc} genome was discarded.")
    elif not qc:
        logger.info(f"{nb_disc} genomes were discarded.")
    # Get input list file name (without path)
    _, name_lst = os.path.split(list_file)
    # if not QC, write discarded genomes to a file "discarded-[list_file].lst"
    if not qc:
        outdisc = os.path.join(res_path,
                               "discarded-" + ".".join(name_lst.split(".")[:-1]) + ".lst")
        logger.info("Writing discarded genomes to {}".format(outdisc))
    # if QC, there is no 'discarded genome', just write information on all analyzed genomes
    else:
        outdisc = os.path.join(res_path,
                               "ALL-GENOMES-info-" + ".".join(name_lst.split(".")[:-1]) + ".lst")
        logger.info("Writing information on genomes in {}".format(outdisc))
    with open(outdisc, "w") as outdf:
        outdf.write("\t".join(["orig_name", "to_annotate", "gsize", "nb_conts", "L90"]) + "\n")
        for genome, values in genomes.items():
            if genome in kept_genomes:
                continue
            _, _, to_annotate, gsize, nbcont, l90 = [str(x) for x in values]
            to_annotate_file = os.path.basename(to_annotate)
            outdf.write("\t".join([genome, to_annotate_file, gsize, nbcont, l90]) + "\n")


def write_lstinfo(list_file, genomes, outdir):
    """
    Write lstinfo file, with following columns:
    gembase_name, orig_name, to_annotate_name, size, nbcontigs, l90

    Parameters
    ----------
    list_file : str
        input file containing the list of genomes
    genomes : dict
        {genome: [gembase_start_name, seq_file, seq_to_annotate, genome_size, nb_contigs, L90]}
    outdir : str
        folder where results must be saved

    """
    _, name_lst = os.path.split(list_file)
    outlst = os.path.join(outdir, "LSTINFO-" + ".".join(name_lst.split(".")[:-1]) + ".lst")
    with open(outlst, "w") as outf:
        outf.write("\t".join(["gembase_name", "orig_name", "to_annotate", "gsize",
                              "nb_conts", "L90"]) + "\n")
        for genome, values in sorted(genomes.items(), key=sort_genomes_byname_l90_nbcont):
            gembase, _, to_annote, gsize, nbcont, l90 = [str(x) for x in values]
            outf.write("\t".join([gembase, genome, to_annote, gsize, nbcont, l90]) + "\n")
    return outlst


def sort_genomes_by_name(x):
    """
    order by:

        - species
        - in each species, by strain number

    Parameters
    ----------
    x : tuple or str
        [genome_orig, [gembase, path, gsize, nbcont, L90]] with gembase = species.date.strain

    Returns
    -------
    str
        variable to take into account for sorting. If format is ESCO.1512.00001 return\
        ESCO and 00001. Otherwise, just return x itself (sort by alphabetical order)
    """
    # get gembase name
    if isinstance(x, tuple):
        x = x[1][0]

    # if format is ESCO.1512.00001 sort by ESCO, then 00001
    if "." in x and len(x.split(".")) >= 3:
        return x.split(".")[0], int(x.split(".")[-1])
    # if format is not like this, just return alphabetical order
    return x,


def sort_genomes_byname_l90_nbcont(x):
    """
    Sort all genomes with the following criteria:

    - sort by species (x[1][0] is species.date)
    - for each species, sort by l90
    - for same l90, sort by nb contigs

    Parameters
    ----------
    x : [[]]
        [genome_name, [species.date, path, path_to_seq, gsize, nbcont, L90]]

    Returns
    -------
    tuple
        information on species, l90 and nb_contigs
    """
    return x[1][0].split(".")[0], x[1][-1], x[1][-2]


def sort_genomes_l90_nbcont(x):
    """
    Sort all genomes with the following criteria:

    - for each strain, sort by l90
    - for same l90, sort by nb contigs

    Parameters
    ----------
    x : [[]]
        [genome_name, [species.date, path, gsize, nbcont, L90]]

    Returns
    -------
    tuple
        information on l90 and nb_contigs
    """
    return x[1][-1], x[1][-2]


def sort_proteins(x):
    """
    order by:

    - species
    - in each species, strain number
    - in each species and strain number, by protein number

    Parameters
    ----------
    x : str
        species.date.strain.contig_protnum

    Returns
    -------
    str
        variable to take into account for sorting. If format is ESCO.1512.00001.i0002_12124,\
        return ESCO, 00001 and 12124. If not, it must be something_00001:\
        return something and 00001.
    """
    try:
        # if format is ESCO.1512.00001.i0002_12124, sort by ESCO, then 00001, then 12124
        if "." in x and len(x.split(".")) >= 3:
            return x.split(".")[0], int(x.split(".")[2].split("_")[0]), int(x.split("_")[-1])
        # if format is not like this, it must be something_00001:
        # sort by 'something' and then 00001
        return "_".join(x.split("_")[:-1]), int(x.split("_")[-1])
    except (IndexError, ValueError):
        logger = logging.getLogger("utils")
        logger.error(("ERROR: Protein {} does not have the required format. "
                      "It must contain, at least <alpha-num>_<num_only>, and at best "
                      "<name>.<date>.<strain_num>.<contig_info>_<prot_num>. "
                      "Please change its name.").format(x))
        sys.exit(1)


def read_genomes(list_file, name, date, dbpath, tmp_path, logger):
    """
    Read list of genomes, and return them.
    If a genome has a name, also return it. Otherwise, return the name given by user.

    Check that the given genome file exists in dbpath. Otherwise, put an error message,
    and ignore this file.

    Parameters
    ----------
    list_file : str
        input file containing the list of genomes
    name : str
        Default species name
    date : str
        Default date
    dbpath : str
        path to folder containing original genome files
    tmp_path : str
        path to folder which will contain the genome files to use before annotation, if\
        needed to change them from original file (for example, merging several contig files\
        in one file, split at each stretch of 5 'N', etc.).

    Returns
    -------
    dict
        {genome: spegenus.date} spegenus.date = name.date
    """
    logger.info("Reading genomes")
    genomes = {}
    # Check that given list file exists
    if not os.path.isfile(list_file):
        logger.error(("ERROR: Your list file '{}' does not exist. "
                      "Please provide a list file.\n Ending program.").format(list_file))
        sys.exit(1)
    # List file exists: open it and read it
    with open(list_file, "r") as lff:
        for line in lff:
            line = line.strip()
            # empty line: go to the next one
            if line == "":
                continue
            # If separator ::, look for species and/or date
            if "::" in line:
                # genomes_inf = genome filename(s) separated by space
                # name_inf = <name>.<date>
                genomes_inf, name_inf = line.split("::")
                genomes_inf = genomes_inf.strip()
                cur_name, cur_date = read_info(name_inf, name, date, genomes_inf)
            # If no separator '::', no information on name and date of genome: use default ones
            # Line only contains genome name (filename in given db_path)
            else:
                genomes_inf = line.strip()
                cur_name = name
                cur_date = date
            # If several file names, check that each one exists, and concatenate the existing files
            genomes_inf = genomes_inf.split()
            if len(genomes_inf) > 1:
                to_concat = []
                for file in genomes_inf:
                    if os.path.isfile(os.path.join(dbpath, file)):
                        to_concat.append(file)
                    else:
                        logger.warning(("{} genome file does not exist. Its file will be "
                                        "ignored when concatenating {}").format(file, genomes_inf))
                # If there are files to concatenate, concatenate them
                if to_concat:
                    genome_name = to_concat[0] + "-all.fna"
                    concat_file = os.path.join(tmp_path, genome_name)
                    to_concat = [os.path.join(dbpath, gname) for gname in to_concat]
                    # Put all genomes listed in 'to_concat' into a same file named 'concat_file'
                    cat(to_concat, concat_file)
                else:
                    # No genome file listed exists. No sequence = genome ignored
                    logger.warning(("None of the genome files in {} exist. "
                                    "This genome will be ignored.").format(genomes_inf))
                    genome_name = ""
            # If only 1 sequence file, check that it exists, and take its name
            else:
                # Genome file given does not exist
                if not os.path.isfile(os.path.join(dbpath, genomes_inf[0])):
                    logger.warning(("{} genome file does not exist. "
                                    "It will be ignored.").format(genomes_inf[0]))
                    genome_name = ""
                # Genome file exists: get genome_name (from information in first field)
                else:
                    genome_name = genomes_inf[0]
            # If there is a genome file (concatenated from several, or already existing), get the full name (with date)
            if genome_name != "":
                genomes[genome_name] = [cur_name + "." + cur_date]
    return genomes


def read_genomes_info(list_file, name, date=None, logger=None):
    """
    Read a lstinfo file containing the list of genomes with information (L90, genome size etc.).
    1 line per genome, 4 required columns (Others will just be ignored):
    to_annotate gsize nb_conts L90

    Check that the given genome file (to_annotate column) exists.

    Parameters
    ----------
    list_file : str
        input file containing information on genomes (to_annotate, size, L90, nb_contigs)
    name : str
        Default species name
    date : str
        Default date
    logger : logging.Logger
        logger object to write log information


    Returns
    -------
    dict
        genomes = {genome:
                   [spegenus.date, path_orig_seq, path_to_splitSequence, size, nbcont, l90]
                  }
    """
    if not logger:
        logger = logging.getLogger("prepare.utils")
    logger.info(f"Reading given information on your genomes in {list_file}")
    genomes = {}
    if name and date:
        spegenus = f"{name}.{date}"
    column_order = {} # Put the number of column corresponding to each field
    if not os.path.isfile(list_file):
        logger.error(f"ERROR: The info file {list_file} that you gave does not exist. "
                      "Please provide the right path/name for this file.\nEnding program.")
        sys.exit(1)
    message_no_header = (f"ERROR: It seems that your info file {list_file} does not have a "
                          "header, or this header does not have, at least, the required "
                          "columns tab separated: to_annotate, gsize nb_conts and L90 (in any "
                          "order).\nEnding program.")
    with open(list_file, "r") as lff:
        for line in lff:
            line = line.strip()
            # Ignore empty lines
            if line == "":
                continue
            # Header line: Just get column number corresponding to each field
            if "to_annotate" in line:
                column_headers = line.split("\t")
                column_order = {header:num for num,header in enumerate(column_headers)}
                found = [head for head in ["to_annotate", "gsize", "nb_conts", "L90"]
                         if head in column_order]
                if len(found) != 4:
                    logger.error(message_no_header)
                    sys.exit(1)
            # If no header found, error message and exit
            if not column_order:
                logger.error(message_no_header)
                sys.exit(1)
            # Get all information on the given genome
            # line.strip().split() -> all given information.
            # So, at least to_annotate, gsize, nbcont, L90
            # Get numeric information
            try:
                infos = line.strip().split()
                # Get genome name with its path to db_dir
                gpath = infos[column_order["to_annotate"]]
                gfile = os.path.basename(gpath)
                gname = os.path.splitext(gfile)[0]
                gsize = int(infos[column_order["gsize"]])
                gl90 = int(infos[column_order["L90"]])
                gcont = int(infos[column_order["nb_conts"]])
            # If invalid values, warning message and ignore genome
            except ValueError:
                logger.warning(f"For genome {gname}, at least one of your columns 'gsize', "
                                "'nb_conts' or 'L90' contains a non numeric value. "
                                "This genome will be ignored.")
                continue
            # If no value for at least 1 field, warning message and ignore genome
            except IndexError:
                logger.error(f"ERROR: Check that all fields of {list_file} are filled in each "
                             "line (can be 'NA')")
                sys.exit(1)
            # Could we find genome file?
            # Check if genome file exists in db_path.
            if not os.path.isfile(gpath):
                logger.warning(f"{gpath} genome file does not exist. This genome will be ignored.")
                continue
            # cur genome information to save:
            # [spegenus.date, path_orig_seq, path_to_sequence_to_annotate, size, nbcont, l90]
            if name and date:
                genomes[gpath] = [spegenus, gpath, gpath, gsize, gcont, gl90]
            # If called from prepare, no need to rename genomes
            else:
                gfile = os.path.basename(gpath)
                gname = os.path.splitext(gfile)[0]
                genomes[gfile] = [gname, gpath, gpath, gsize, gcont, gl90]
    if len(genomes) > 0:
        logger.info(("Found {} genomes in total").format(len(genomes)))
    else:
        logger.error(f"No genome listed in {list_file} was found.")
        sys.exit(1)
    return genomes


def read_info(name_inf, name, date, genomes_inf):
    """
    From the given information in 'name_inf', check if there is a name (and if its
    format is ok) and if there is a date (and if its format is ok).
    If no name (resp. no date), return default name (resp. default date).

    Parameters
    ----------
    name_inf : str
        information on current genome, which could contain a species name and a date
    name : str
        default species name
    date : str
        default date
    genomes_inf : str
        current genome filename. Used to complete information when there is a warning (species\
        name or date given not in the right format...)

    Returns
    -------
    (cur_name, cur_date) : tuple
        with:

        - curname: name to use for this genome (can be the default one, or the one read from\
        'name_inf'
        - curdate: date to use for this genome (default or read from 'name_inf')
    """
    logger = logging.getLogger("utils")
    # information on current genome, which could contain a species name and a date: split to get
    # name and date
    name_inf = name_inf.strip().split(".")
    # if only 1 field provided (means only name is provided).
    # (Otherwise, if only date is provided, it is with '.DATE', so
    # len(name_inf) = 2 with name_inf[0] = "")
    if len(name_inf) == 1:
        # No genome name: use default one
        if name_inf[0] == "":
            cur_name = name
        # Genome name, and in right format: use it
        elif check_format(name_inf[0]):
            cur_name = name_inf[0]
        # Genome name but wrong format: warning message and use default one
        else:
            logger.warning(("Invalid name {} given for genome {}. Only put "
                            "4 alphanumeric characters in your date and name. "
                            "For this genome, the default name ({}) will be "
                            "used.").format(name_inf[0], genomes_inf, name))
            cur_name = name
        # In any case, no given date: use default date
        cur_date = date

    # More than 2 informations (more than 2 fields separated by a '.'):
    #  - either user put more information
    #  - either user put a '.' inside the name or date field: wrong format
    # -> warning message and use default date and name.
    elif len(name_inf) > 2:
        logger.warning(("Invalid name/date given for genome {}. Only put "
                        "4 alphanumeric characters in your date and name. For "
                        "this genome, the default name ({}) and date ({}) will "
                        "be used.").format(genomes_inf, name, date))
        cur_name = name
        cur_date = date
    # information on name and date given (can be empty)
    else:
        cur_name, cur_date = name_inf
        # name given empty -> use default name
        if cur_name == "":
            cur_name = name
        # date given empty -> use default date
        if cur_date == "":
            cur_date = date
        # name given not empty but wrong format -> warning message and default name
        if not check_format(cur_name):
            logger.warning(("Invalid name {} given for genome {}. Only put "
                            "4 alphanumeric characters in your date and name. "
                            "For this genome, the default name ({}) "
                            "will be used.").format(cur_name, genomes_inf, name))
            cur_name = name
        # date given not empty but wrong format -> warning message and default date
        if not check_format(cur_date):
            logger.warning(("Invalid date {} given for genome {}. Only put "
                            "4 alphanumeric characters in your date and name. "
                            "For this genome, the default date ({}) "
                            "will be used.").format(cur_date, genomes_inf, date))
            cur_date = date
    return cur_name, cur_date


def cat(list_files, output, title=None):
    """
    Equivalent of 'cat' unix command.

    Concatenate all files in 'list_files' and save result in 'output' folder.
    Concat using shutil.copyfileobj, in order to copy by chunks, to
    avoid memory problems if files are big.

    Parameters
    ----------
    list_files : list
        list of filenames to concatenate
    output : str
        output filename, where all concatenated files will be written
    title : str or None
        if you want to show a progressbar while concatenating files, add a title for this
        progressbar here. If no title, nothing will be shown during concatenation.

    """
    bar = None
    curnum = None
    if title:
        nbfiles = len(list_files)
        widgets = [title + ': ', progressbar.Bar(marker='█', left='', right='', fill=' '),
                   ' ', progressbar.Counter(), f"/{nbfiles}" ' (',
                   progressbar.Percentage(), ") - ", progressbar.Timer()]
        bar = progressbar.ProgressBar(widgets=widgets, max_value=nbfiles, term_width=79).start()
        curnum = 1
    with open(output, "w") as outf:
        for file in list_files:
            if title:
                bar.update(curnum)
                curnum += 1
            with open(file, "r") as inf:
                shutil.copyfileobj(inf, outf)
    if title:
        bar.finish()


def grep(filein, pattern, counts=False):
    """
    Equivalent of 'grep' unix command

    By default, returns all the lines containing the given pattern.
    If counts = True, returns the number of lines containing the pattern.

    Parameters
    ----------
    filein : str
        path to the file in which pattern must be searched
    pattern : str
        pattern to search
    counts : bool
        True if you want to count how many lines have the pattern and return this number,\
        False if you want to return all lines containing the pattern.

    Returns
    -------
    list or int
        list of lines if counts=False; number of lines if counts=True
    """
    num = 0
    lines = []
    with open(filein, "r") as inf:
        for line in inf:
            if re.search(pattern, line):
                lines.append(line.strip())
                num += 1
    if counts:
        return num
    else:
        return lines


def count(filein, get="lines"):
    """
    Similar to 'wc' unix command.

    Count the number of what is given in 'get'. It can be:

    - lines (default)
    - words

    Parameters
    ----------
    filein : str
        path to the file for which we want to count lines or words
    get : ["lines", "words"]
        either lines to count the number of lines in the file, or words to count the number\
        of words.

    Returns
    -------
    int
        Number of lines or words according to value of 'get' parameter.
    """
    gets = ["lines", "words"]
    if get not in gets:
        logger = logging.getLogger("utils")
        logger.error("Choose what you want to count among {}.".format(gets))
        sys.exit(1)
    num = 0
    with open(filein, "r") as inf:
        for line in inf:
            if get == "lines":
                num += 1
            elif get == "words":
                num += len(line.split())
    return num


def check_format(info):
    """
    Check that the given information (can be the genomes name or the date) is in the right
    format: it should have 4 characters, all alphanumeric.

    Parameters
    ----------
    info : str
        information to check

    Returns
    -------
    bool
        True if right format, False otherwise
    """
    if len(info) != 4:
        return False
    return info.isalnum()


def check_out_dirs(resdir):
    """
    Check that there is no file in:

    - resdir/LSTINFO
    - resdir/Genes
    - resdir/Proteins
    - resdir/Replicons
    - resdir/gff3

    Parameters
    ----------
    resdir : str
        path to result directory

    """
    logger = logging.getLogger("utils")
    if glob.glob(os.path.join(resdir, "LSTINFO", "*.lst")):
        logger.error("ERROR: Your output directory already has .lst files in the "
                     "LSTINFO folder. Provide another result directory, or remove the "
                     "files in this one.\nEnding program.")
        sys.exit(1)
    if glob.glob(os.path.join(resdir, "Proteins", "*.prt")):
        logger.error("ERROR: Your output directory already has .prt files in the "
                     "Proteins folder. Provide another result directory, or remove the "
                     "files in this one.\nEnding program.")
        sys.exit(1)
    if glob.glob(os.path.join(resdir, "Genes", "*.gen")):
        logger.error("ERROR: Your output directory already has .gen files in the "
                     "Genes folder. Provide another result directory, or remove the "
                     "files in this one.\nEnding program.")
        sys.exit(1)
    if glob.glob(os.path.join(resdir, "Replicons", "*.fna")):
        logger.error("ERROR: Your output directory already has .fna files in the "
                     "Replicons folder. Provide another result directory, or remove the "
                     "files in this one.\nEnding program.")
        sys.exit(1)
    if glob.glob(os.path.join(resdir, "gff3", "*.gff")):
        logger.error("ERROR: Your output directory already has .gff files in the "
                     "gff3 folder. Provide another result directory, or remove the "
                     "files in this one.\nEnding program.")
        sys.exit(1)


def get_genome_contigs_and_rename(gembase_name, gpath, outfile, logger):
    """
    For the given genome (sequence in gpath), rename all its contigs
    with the new name: 'gembase_name', and save the output sequence in outfile.

    For each contig renamed, save its new name as well as its size. This will be used to generate
    gff files

    Parameters
    ----------
    gembase_name : str
        genome name to use (species.date.strain)
    gpath : str
        path to the genome sequence
    outfile : str
        path to the new file, containing 'gpath' sequence, but with 'gembase_name' in headers

    Returns
    -------
    tuple
        - Dict of all contigs with their original and new name: (list of str)
        {>orig_name: >new_name}
        - Dict of all contigs with their size: (list of str)
        {"new_name': 'size1"}
    """
    # Initialize variables

    # Contig number
    contig_num = 1
    # contig size
    cont_size = 0
    # List of contigs (str) [<name>\t<orig_name>]
    contigs = {}
    # List of contigs (str) with their sizes [<name>\t<size>]
    sizes = {}
    # Name of previous contig (to put to contigs, as we need to wait for the next
    # contig to know the size of the previous one)
    prev_cont = ""
    prev_orig_name = ""
    # sequence of previous contig
    seq = ""

    # Read input sequence given to prodigal, and open file where sequences with new
    # headers must be written.
    with open(gpath, "r") as gpf, open(outfile, "w") as grf:
        for line in gpf:
            # When we find a new header line, convert its name to gembase format, and write it
            # to output replicon file
            if line.startswith(">") :
                # If not first contig (contigs not empty):
                # - add its name as well as its size to contigs list
                # - add its name with its original name to
                # - write header ("<contig name> <size>") to replicon file
                if prev_cont:
                    cont = "\t".join([prev_cont, str(cont_size)]) + "\n"
                    prevcont_nohead = prev_cont.split(">")[1]
                    prev_orig_name_nohead = prev_orig_name.split(">")[1]
                    if prev_orig_name_nohead:
                        if prev_orig_name_nohead in contigs:
                            logger.error(f"several contigs have the same name "
                                         f"{prev_orig_name_nohead} in {gpath}.")
                            return False, False
                        sizes[prevcont_nohead] = cont_size
                        contigs[prev_orig_name_nohead] = prevcont_nohead
                        grf.write(cont)
                        grf.write(seq)
                prev_cont = ">" + gembase_name + "." + str(contig_num).zfill(4)
                # keep only first string of contig
                prev_orig_name = line.strip().split()[0]
                contig_num += 1
                cont_size = 0
                seq = ""
            # Sequence line: write it as is in replicon file, and add its size to cont_size.
            else:
                seq += line
                cont_size += len(line.strip())
        # Write last contig, if there is one (if gpath not empty)
        if prev_cont:
            cont = "\t".join([prev_cont, str(cont_size)]) + "\n"
            prevcont_nohead = "".join(prev_cont.split(">")[1:])
            prev_orig_name_nohead = prev_orig_name.split(">")[1]
            if prev_orig_name_nohead:
                if prev_orig_name_nohead in contigs:
                    logger.error(f"several contigs have the same name {prev_orig_name_nohead} "
                                 f"in {gpath}.")
                    return False, False
            contigs[prev_orig_name_nohead] = prevcont_nohead
            sizes[prevcont_nohead] = cont_size
            grf.write(cont)
            grf.write(seq)
    if not contigs:
        logger.error(f"Your genome {gpath} does not contain any sequence, "
                     "or is not in fasta format.")
    return contigs, sizes
# Add test with empty gpath
# Add test with non fasta gpath


def logger_thread(q):
    """
    Queue listener used in a thread to handle the logs put to a QueueHandler
    by several processes (multiprocessing.pool.map_async for example)

    Parameters
    ----------
    q : multiprocessing.managers.AutoProxy[Queue]
        queue to listen

    """
    while True:
        record = q.get()
        if record is None:
            break
        logger = logging.getLogger(record.name)
        logger.handle(record)


def detail_lvl():
    """
    Get the int level corresponding to "DETAIL"

    Returns
    -------
    int
        int corresponding to the level "DETAIL"
    """
    return 15


def save_bin(objects, fileout):
    """
    Save python 'objects' in a binary file called 'fileout'

    Parameters
    ----------
    objects : Object
        python object to save
    fileout : str
        path to binary file where objects must be saved

    """
    with open(fileout, "wb") as binf:
        pickle.dump(objects, binf)


def load_bin(binfile):
    """
    Unpickle python objects from the binary file 'binfile'

    Parameters
    ----------
    binfile : str
        path to binary file containing python object

    Returns
    -------
    Object
        The python objects unpickled

    """
    with open(binfile, "rb") as binf:
        objects = pickle.load(binf)
    return objects


def write_list(list_names, fileout):
    """
    Write the given list of strings to a file, 1 per line
    """
    with open(fileout, "w") as fo:
        for genome in list_names:
            fo.write(str(genome) + "\n")


def list_to_str(list, sep='\t'):
    """
    Return a string corresponding to the given list, with all elements separated
    by a space. Used to write a list into a file. Ex::

        [1, 2, "toto"] -> "1 2 toto"

    Parameters
    ----------
    list : list
        list of elements that we would like to write
    sep : str
        Separator to use between the different elements

    Returns
    -------
    str
        the string to write
    """
    list_write = [str(l) for l in list]
    return sep.join(list_write) + "\n"


def remove(infile):
    """
    Remove the given file if it exists

    Parameters
    ----------
    infile : str
        path to file to remove

    """
    if os.path.isfile(infile):
        os.remove(infile)



#!/usr/bin/env python3
# coding: utf-8

# ###############################################################################
# This file is part of PanACOTA.                                                #
#                                                                               #
# Authors: Amandine Perrin                                                      #
# Copyright © 2018-2020 Institut Pasteur (Paris).                               #
# See the COPYRIGHT file for details.                                           #
#                                                                               #
# PanACOTA is a software providing tools for large scale bacterial comparative  #
# genomics. From a set of complete and/or draft genomes, you can:               #
#    -  Do a quality control of your strains, to eliminate poor quality         #
# genomes, which would not give any information for the comparative study       #
#    -  Uniformly annotate all genomes                                          #
#    -  Do a Pan-genome                                                         #
#    -  Do a Core or Persistent genome                                          #
#    -  Align all Core/Persistent families                                      #
#    -  Infer a phylogenetic tree from the Core/Persistent families             #
#                                                                               #
# PanACOTA is free software: you can redistribute it and/or modify it under the #
# terms of the Affero GNU General Public License as published by the Free       #
# Software Foundation, either version 3 of the License, or (at your option)     #
# any later version.                                                            #
#                                                                               #
# PanACOTA is distributed in the hope that it will be useful, but WITHOUT ANY   #
# WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS     #
# FOR A PARTICULAR PURPOSE. See the Affero GNU General Public License           #
# for more details.                                                             #
#                                                                               #
# You should have received a copy of the Affero GNU General Public License      #
# along with PanACOTA (COPYING file).                                           #
# If not, see <https://www.gnu.org/licenses/>.                                  #
# ###############################################################################

"""
Functions to infer a phylogenetic tree with fastME

@author gem
June 2017
"""

from Bio import AlignIO
import os
import logging

from PanACoTA import utils

logger = logging.getLogger("tree.fastme")

def run_tree(alignfile, boot, outdir, quiet, threads, **kwargs):
    """
    Run fastme for the given alignment file and options

    Parameters
    ----------
    alignfile: str
        Path to file containing alignments of persistent families grouped by genome
    boot: int or None
        Number of bootstraps to compute. None if no bootstrap asked
    outdir: str
        output directory to save all results
    quiet: bool
        True if nothing must be printed to stderr/stdout, False otherwise
    threads: int
        Maximum number of threads to use
    kwargs["model"]: str
        DNA substitution model chosen by user
    kwargs["wb"]: bool
        True if all bootstrap pseudo-trees must be saved into a file, False otherwise
    """
    model = kwargs["model"]
    write_boot = kwargs["wb"]
    align_name = os.path.basename(alignfile)
    align_phylip = os.path.join(outdir, align_name + ".phylip")
    convert2phylip(alignfile, align_phylip)
    run_fastme(align_phylip, boot, write_boot, threads, model, outdir, quiet)


def convert2phylip(infile, outfile):
    """
    Input alignment is in fasta format. Input of fastME must be in Phylip-relaxed format.
    Convert it here.

    Parameters
    ----------
    infile: str
        Path to file in fasta format
    outfile: str
        Path to file to generate, in Phylip-relaxed format
    """
    if os.path.isfile(outfile):
        logger.info("Phylip alignment file already existing.")
        logger.warning(("The Phylip alignment file {} already exists. The program "
                        "will use it instead of re-converting {}.").format(outfile, infile))
        return
    logger.info("Converting fasta alignment to PHYLIP-relaxed format.")
    with open(infile, 'r') as input_handle, open(outfile, 'w') as output_handle:
        alignments = AlignIO.parse(input_handle, "fasta")
        AlignIO.write(alignments, output_handle, "phylip-relaxed")


def run_fastme(alignfile, boot, write_boot, threads, model, outdir, quiet):
    """
    Run fastME on the given alignment.

    Parameters
    ----------
    alignfile: str
        Path to file containing alignments of persistent families grouped by genome
    boot: int or None
        Number of bootstraps to compute. None if no bootstrap asked
    write_boot: bool
        True if all bootstrap pseudo-trees must be saved into a file, False otherwise
    threads: int
        Maximum number of threads to use
    model: str or None
        DNA substitution model chosen by user. None if default one
    outdir: str
        output directory to save all results
    quiet: bool
        True if nothing must be printed to stderr/stdout, False otherwise
    """
    logger.info("Running FastME...")
    bootinfo = ""
    threadinfo = ""
    outboot = ""

    # Get bootstrap information
    if boot:
        bootinfo = "-b {}".format(boot)
    # Get threads information
    if threads:
        threadinfo = "-T {}".format(threads)
    # Get output filename
    align_name = os.path.basename(alignfile)
    logfile = os.path.join(outdir, align_name + ".fastme.log")
    treefile = os.path.join(outdir, align_name + ".fastme_tree.nwk")
    # If bootstrap pseudo-trees must be written, define the filename here
    if write_boot:
        outboot = "-B " + os.path.join(outdir, align_name + ".fastme_bootstraps.nwk")
    # Put default model if not given
    if not model:
        model = "T"
    cmd = (f"fastme -i {alignfile} -d{model} -nB -s {threadinfo} {bootinfo} "
           f"-o {treefile} -I {logfile} {outboot}")
    logger.details(cmd)
    if quiet:
        fnull = open(os.devnull, 'w')
    else:
        fnull = None
    error = ("Problem while running FastME. See log file ({}) for "
             "more information.").format(logfile)
    utils.run_cmd(cmd, error, stdout=fnull, eof=True, logger=logger, stderr=fnull)



#!/usr/bin/env python3
# coding: utf-8

"""
Unit tests for the protein_seq_functions in pangenome module
"""

import os
import shutil
import logging
import pytest

import PanACoTA.pangenome_module.protein_seq_functions as psf
import test.test_unit.utilities_for_tests as tutil


# Define variables and functions used by several tests
PANDIR = os.path.join("test", "data", "pangenome")
PATH_TEST_FILES = os.path.join(PANDIR, "test_files")
PATH_EXP_FILES = os.path.join(PANDIR, "exp_files")
GENEPATH = os.path.join(PANDIR, "generated_by_unit-tests")


@pytest.fixture(autouse=True)
def setup_teardown_module():
    """
    Remove log files at the end of this test module

    Before each test:
    - init logger
    - create directory to put generated files

    After:
    - remove all log files
    - remove directory with generated results
    """
    os.mkdir(GENEPATH)
    print("setup")

    yield
    shutil.rmtree(GENEPATH)
    print("teardown")



# Start tests
def test_build_bank(caplog):
    """
    Build a protein bank from a list of genomes, and create it at the same
    place as the database.
    """
    caplog.set_level(logging.DEBUG)
    lstinfo = os.path.join(PATH_TEST_FILES, "list_to_pan.txt")
    dbpath = os.path.join(PATH_TEST_FILES, "example_db", "Proteins")
    cur_dbpath = os.path.join(GENEPATH, "Proteins")
    shutil.copytree(dbpath, cur_dbpath)
    name = "EXEM"
    spedir = None
    quiet = True
    outfile = psf.build_prt_bank(lstinfo, cur_dbpath, name, spedir, quiet)
    exp_file = os.path.join(PATH_EXP_FILES, "exp_EXEM.All.prt")
    exp_out = os.path.join(cur_dbpath, name + ".All.prt")

    # Check prt bank filename
    assert outfile == exp_out
    # Check content of bank created
    assert tutil.compare_order_content(exp_file, exp_out)

    # Check logs
    assert ("Building bank with all proteins to test/data/pangenome/"
            "generated_by_unit-tests/Proteins/EXEM.All.prt") in caplog.text


def test_build_bank_spedir_quiet(caplog):
    """
    Build a protein bank from a list of genomes, and create it in a given output directory.
    """
    caplog.set_level(logging.DEBUG)
    lstinfo = os.path.join(PATH_TEST_FILES, "list_to_pan.txt")
    dbpath = os.path.join(PATH_TEST_FILES, "example_db", "Proteins")
    name = "EXEM"
    spedir = os.path.join(GENEPATH, "test_build_prt", "toto")
    quiet = False
    outfile = psf.build_prt_bank(lstinfo, dbpath, name, spedir, quiet)
    exp_file = os.path.join(PATH_EXP_FILES, "exp_EXEM.All.prt")
    exp_out = os.path.join(spedir, name + ".All.prt")
    assert outfile == exp_out
    assert tutil.compare_order_content(exp_file, exp_out)
    # Check logs
    assert ("Building bank with all proteins to test/data/pangenome/"
            "generated_by_unit-tests/test_build_prt/toto/EXEM.All.prt") in caplog.text


def test_build_bank_exists(caplog):
    """
    Test that when we want to create a bank but the output file already exists, it prints
    a warning, and closes without redoing the bank
    """
    caplog.set_level(logging.DEBUG)
    lstinfo = os.path.join(PATH_TEST_FILES, "list_to_pan.txt")
    dbpath = os.path.join(PATH_TEST_FILES, "example_db", "Proteins")
    cur_dbpath = os.path.join(GENEPATH, "Proteins")
    shutil.copytree(dbpath, cur_dbpath)
    name = "EXEM"
    spedir = None
    exp_out = os.path.join(cur_dbpath, name + ".All.prt")
    exp_file = os.path.join(PATH_EXP_FILES, "exp_EXEM.All.prt")
    shutil.copyfile(exp_file, exp_out)
    quiet = True
    outfile = psf.build_prt_bank(lstinfo, cur_dbpath, name, spedir, quiet)
    assert outfile == exp_out
    assert ("Protein bank test/data/pangenome/generated_by_unit-tests/Proteins/"
            "EXEM.All.prt already exists. It will be used by mmseqs.") in caplog.text
    assert caplog.records[0].levelname == "WARNING"



#!/usr/bin/env python3
# coding: utf-8

"""
Tests for make script, installing PanACoTA according to already existing dependencies
From a computer with ubuntu, prokka and quicktree installed
"""
import os
import pytest
import glob

from . import utilities as utils

@pytest.fixture
def install_panacota():
    cmd = "python3 make"
    error = "Error installing"
    utils.run_cmd(cmd, error)


def teardown_function(function):
    """
    Uninstall PanACoTA and installed dependencies at the end of each test
    """
    print("TEARDOWN\n")
    cmd = "python3 make uninstall"
    error = "Error uninstall"
    utils.run_cmd(cmd, error)
    os.remove("install.log")
    print("cleaning repo")


def test_install_prokka_quicktree():
    """
    Test that when installing from a computer containing only prokka and quicktree, it installs
    PanACoTA, and returns the list of missing dependencies
    """
    assert not utils.check_installed("barrnap")
    assert not utils.check_installed("prodigal")
    assert utils.check_installed("quicktree")
    assert not utils.check_installed("iqtree")
    assert utils.check_installed("prokka")
    assert not utils.check_installed("PanACoTA")
    assert utils.check_installed('quicktree')
    cmd = "python3 make"
    error = "Error trying to install PanACoTA from base"
    utils.run_cmd(cmd, error)
    assert not utils.check_installed("barrnap")
    assert utils.check_installed("prokka")
    assert utils.check_installed("PanACoTA")
    assert utils.check_installed('quicktree')
    cmd = "pip3 show PanACoTA"
    err = "error pip3"
    stdout = "stdout_pip3show.out"
    with open(stdout, "w") as stdof:
        utils.run_cmd(cmd, err, stdout=stdof, stderr=stdof)
    with open(stdout, "r") as stdof:
        for line in stdof:
            if line.startswith("Location"):
                loc = line.split()[-1]
                assert glob.glob(os.path.join(loc, r'PanACoTA*dist-info'))
    os.remove(stdout)
    logfile = "install.log"
    content = ["Installing PanACoTA...", "DONE",
               "Some dependencies needed for some subcommands of PanACoTA are "
               "not installed. Here is the list of missing dependencies, and for what they are "
               "used. If you plan to use the subcommands hereafter, first install required "
               "dependencies:",
               "- mmseqs (for pangenome subcommand)",
               "- mash (for prepare subcommand, to filter genomes)",
               "- mafft (to align persistent genomes in order to infer a phylogenetic tree "
               "after)",
               "See more information on how to download/install those softwares in README or in "
               "documentation.",
               "- barrnap. If you use Prokka for functional annotation, it will not predict RNA."
               ]
    with open(logfile, "r") as logf:
        logf_content = "".join(logf.readlines())
        for linec in content:
            assert linec in logf_content
    # Check that needed packages are installed
    assert utils.is_package_installed("argparse")
    assert utils.is_package_installed("progressbar")
    assert utils.is_package_installed("numpy")
    assert utils.is_package_installed("matplotlib")
    assert utils.is_package_installed("Bio")
    os.remove(logfile)


def test_upgrade(install_panacota):
    """
    Test upgrading PanACoTA when dependencies are still installed
    """
    assert not utils.check_installed("barrnap")
    assert utils.check_installed("prokka")
    assert utils.check_installed("PanACoTA")
    assert utils.check_installed('quicktree')
    cmd = "python3 make upgrade"
    error = "Error upgrade"
    utils.run_cmd(cmd, error)
    assert not utils.check_installed("barrnap")
    assert utils.check_installed("prokka")
    assert utils.check_installed("PanACoTA")
    assert utils.check_installed('quicktree')
    logfile = "install.log"
    with open(logfile, "r") as logf:
        lines = logf.readlines()
        assert len(lines) == 2
        assert "Upgrading PanACoTA" in lines[0]
        assert "DONE" in lines[1]


def test_uninstall_withdep(install_panacota):
    """
    Test uninstalling PanACoTA when dependencies are still installed
    """
    assert not utils.check_installed("barrnap")
    assert utils.check_installed("prokka")
    assert utils.check_installed("quicktree")
    assert utils.check_installed("PanACoTA")
    cmd = "python3 make uninstall"
    error = "Error uninstalling"
    utils.run_cmd(cmd, error)
    assert not utils.check_installed("barrnap")
    assert utils.check_installed("prokka")
    assert not utils.check_installed("PanACoTA")
    assert utils.check_installed('quicktree')


def test_develop():
    """
    Test installing PanACoTA in developer mode, when prokka and barrnap are already installed
    """
    assert not utils.check_installed("PanACoTA")
    assert not utils.check_installed("barrnap")
    assert utils.check_installed('quicktree')
    assert utils.check_installed("prokka")
    cmd = "python3 make develop"
    error = "Error develop"
    utils.run_cmd(cmd, error)
    assert not utils.check_installed("barrnap")
    assert utils.check_installed("prokka")
    assert utils.check_installed("PanACoTA")
    assert utils.check_installed('quicktree')
    cmd = "pip3 show PanACoTA"
    err = "error pip3"
    stdout = "stdout_pip3show.out"
    with open(stdout, "w") as stdof:
        utils.run_cmd(cmd, err, stdout=stdof, stderr=stdof)
    # Check installation in develop mode
    with open(stdout, "r") as stdof:
        for line in stdof:
            if line.startswith("Location"):
                loc = line.split()[-1]
                assert glob.glob(os.path.join(loc, r'PanACoTA*egg-info'))
    os.remove(stdout)
    logfile = "install.log"
    content = ["Installing PanACoTA...",
               "Installing developer packages needed for PanACoTA",
               "Some dependencies needed for some subcommands of PanACoTA are "
               "not installed. Here is the list of missing dependencies, and for what they are "
               "used. If you plan to use the subcommands hereafter, first install required "
               "dependencies:",
               "- mmseqs (for pangenome subcommand)",
               "- mafft (to align persistent genomes in order to infer a phylogenetic tree "
               "after)",
               "- mash (for prepare subcommand, to filter genomes)",
               "- barrnap. If you use Prokka for functional annotation, it will not predict RNA.",
               "See more information on how to download/install those softwares in README or in "
               "documentation."]

    with open(logfile, "r") as logf:
        logf_content = "".join(logf.readlines())
        for linec in content:
            assert linec in logf_content

    # Check that needed packages are installed
    assert utils.is_package_installed("argparse")
    assert utils.is_package_installed("progressbar")
    assert utils.is_package_installed("numpy")
    assert utils.is_package_installed("matplotlib")
    assert utils.is_package_installed("Bio")
    os.remove(logfile)


def test_install_user():
    """
    Test that when installing from a computer containing only prokka, in user mode, it installs
    PanACoTA in /Users and returns list of dependencies
    """
    assert not utils.check_installed("barrnap")
    assert not utils.check_installed("mash")
    assert utils.check_installed("prokka")
    assert utils.check_installed("quicktree")
    assert not utils.check_installed("PanACoTA")
    cmd = "python3 make --user"
    error = "Error trying to install PanACoTA from base"
    utils.run_cmd(cmd, error)
    assert not utils.check_installed("barrnap")
    assert not utils.check_installed("mash")
    assert utils.check_installed("prokka")
    assert utils.check_installed("quicktree")
    assert utils.check_installed("PanACoTA")
    # Check logfile content
    logfile = "install.log"
    content = ["Installing PanACoTA in user mode...",
               "Some dependencies needed for some subcommands of PanACoTA are "
               "not installed. Here is the list of missing dependencies, and for what they are "
               "used. If you plan to use the subcommands hereafter, first install required "
               "dependencies:",
               "- mmseqs (for pangenome subcommand)",
               "- mafft (to align persistent genomes in order to infer a phylogenetic tree "
               "after)",
               "- mash (for prepare subcommand, to filter genomes)",
               "See more information on how to download/install those softwares in README or in "
               "documentation."]
    with open(logfile, "r") as logf:
        logf_content = "".join(logf.readlines())
        for linec in content:
            assert linec in logf_content
    # Check that needed packages are installed
    assert utils.is_package_installed("argparse")
    assert utils.is_package_installed("progressbar")
    assert utils.is_package_installed("numpy")
    assert utils.is_package_installed("matplotlib")
    assert utils.is_package_installed("Bio")
